"""
Generated by generate/generate.py - 2020-09-08 17:49:51.525706
"""
from aiogithubapi.objects.login.device import AIOGitHubAPILoginDevice

from tests.legacy.responses.login.device_fixtrue import device_fixtrue_response


def test_device(device_fixtrue_response):
    obj = AIOGitHubAPILoginDevice(device_fixtrue_response)
    assert obj.device_code == device_fixtrue_response["device_code"]
    assert obj.user_code == device_fixtrue_response["user_code"]
    assert obj.verification_uri == device_fixtrue_response["verification_uri"]
    assert obj.expires_in == device_fixtrue_response["expires_in"]
    assert obj.interval == device_fixtrue_response["interval"]



"""Test events namespace."""
# pylint: disable=missing-docstring
import asyncio
from unittest.mock import AsyncMock, patch

import pytest
import pytest_asyncio
import aiohttp

from aiogithubapi import GitHubAPI, GitHubEventModel

from tests.common import TEST_REPOSITORY_NAME, TOKEN, MockedRequests, MockResponse


@pytest_asyncio.fixture
async def wait_mock():
    async def _mocker(_, __):
        await asyncio.sleep(0)

    with patch("aiogithubapi.namespaces.events._GitHubEventsBaseNamespace._wait", _mocker):
        yield


@pytest.mark.asyncio
async def test_subscription(github_api: GitHubAPI, mock_requests: MockedRequests):
    tasks_before = asyncio.all_tasks(
        github_api._client._loop)  # This includes the test task
    assert len(tasks_before) == 1

    event_callback_mock = AsyncMock()

    subscription_id = await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME, event_callback=event_callback_mock
    )

    assert subscription_id in github_api.repos.events._subscriptions
    while not event_callback_mock.called:
        await asyncio.sleep(0)

    event: GitHubEventModel = event_callback_mock.call_args[0][0]
    assert event.type == "PushEvent"

    assert len(asyncio.all_tasks(github_api._client._loop) - tasks_before) == 1

    tasks = github_api.repos.events.unsubscribe(
        subscription_id=subscription_id)
    await asyncio.wait(tasks)
    assert not github_api.repos.events._subscriptions
    assert len(asyncio.all_tasks(github_api._client._loop) - tasks_before) == 0


@pytest.mark.asyncio
async def test_stopping_all_subscriptions(github_api: GitHubAPI, mock_requests: MockedRequests):
    tasks_before = asyncio.all_tasks(
        github_api._client._loop)  # This includes the test task
    assert len(tasks_before) == 1

    event_callback_mock = AsyncMock()

    subscription_id_1 = await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME, event_callback=event_callback_mock
    )
    subscription_id_2 = await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME, event_callback=event_callback_mock
    )

    assert subscription_id_1 in github_api.repos.events._subscriptions
    assert subscription_id_2 in github_api.repos.events._subscriptions
    while not event_callback_mock.called:
        await asyncio.sleep(0)

    event: GitHubEventModel = event_callback_mock.call_args[0][0]
    assert event.type == "PushEvent"

    assert len(asyncio.all_tasks(github_api._client._loop) - tasks_before) == 2

    await github_api.repos.events.unsubscribe_all()
    assert len(asyncio.all_tasks(github_api._client._loop) - tasks_before) == 0


@pytest.mark.asyncio
async def test_stopping_all_subscriptions_async_with(
    client_session: aiohttp.ClientSession,
    mock_requests: MockedRequests,
):
    event_loop = asyncio.get_event_loop()
    tasks_before = asyncio.all_tasks(event_loop)  # This includes the test task
    assert len(tasks_before) == 1

    event_callback_mock = AsyncMock()

    async with GitHubAPI(token=TOKEN, session=client_session) as github_api:
        subscription_id = await github_api.repos.events.subscribe(
            TEST_REPOSITORY_NAME, event_callback=event_callback_mock
        )

        assert subscription_id in github_api.repos.events._subscriptions
        while not event_callback_mock.called:
            await asyncio.sleep(0)

        event: GitHubEventModel = event_callback_mock.call_args[0][0]
        assert event.type == "PushEvent"

        assert len(asyncio.all_tasks(event_loop) - tasks_before) == 1

    assert len(asyncio.all_tasks(event_loop) - tasks_before) == 0


@pytest.mark.asyncio
async def test_subscription_exceptions_not_modified(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    mock_requests: MockedRequests,
    wait_mock: None,
):
    mock_requests.clear()
    event_callback_mock = AsyncMock()
    error_callback_mock = AsyncMock()

    mock_response.mock_status = 304

    subscription_id = await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME,
        event_callback=event_callback_mock,
        error_callback=error_callback_mock,
    )

    while mock_requests.called < 10:
        await asyncio.sleep(0)

    assert subscription_id in github_api.repos.events._subscriptions
    assert not event_callback_mock.called
    assert not error_callback_mock.called

    github_api.repos.events.unsubscribe()
    assert not github_api.repos.events._subscriptions


@pytest.mark.asyncio
async def test_subscription_exceptions_not_found(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    wait_mock: None,
):
    event_callback_mock = AsyncMock()
    error_callback_mock = AsyncMock()

    mock_response.mock_status = 404

    await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME,
        event_callback=event_callback_mock,
        error_callback=error_callback_mock,
    )

    while not error_callback_mock.called:
        await asyncio.sleep(0)

    assert not event_callback_mock.called
    assert error_callback_mock.called

    github_api.repos.events.unsubscribe()
    assert not github_api.repos.events._subscriptions


@pytest.mark.asyncio
async def test_subscription_exception(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    mock_requests: MockedRequests,
    wait_mock: None,
):
    event_callback_mock = AsyncMock()
    error_callback_mock = AsyncMock()

    mock_response.mock_status = 400

    await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME,
        event_callback=event_callback_mock,
        error_callback=error_callback_mock,
    )

    while not error_callback_mock.called:
        await asyncio.sleep(0)

    while mock_requests.called < 10:
        await asyncio.sleep(0)

    assert not event_callback_mock.called
    assert error_callback_mock.called

    github_api.repos.events.unsubscribe()
    assert not github_api.repos.events._subscriptions


@pytest.mark.asyncio
async def test_subscription_exception_in_handler(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    wait_mock: None,
):
    event_callback_mock = AsyncMock(side_effect=TypeError)
    error_callback_mock = AsyncMock()

    await github_api.repos.events.subscribe(
        TEST_REPOSITORY_NAME,
        event_callback=event_callback_mock,
        error_callback=error_callback_mock,
    )

    while not event_callback_mock.called and not error_callback_mock.called:
        await asyncio.sleep(0)

    assert event_callback_mock.called
    assert error_callback_mock.called

    github_api.repos.events.unsubscribe()
    assert not github_api.repos.events._subscriptions



"""
Methods for the repos namespace

https://docs.github.com/en/rest/reference/repos
"""

from __future__ import annotations

from typing import Any, Dict, List

from ..const import GitHubRequestKwarg, RepositoryType
from ..helpers import repository_full_name
from ..models.commit import GitHubCommitModel
from ..models.repository import GitHubRepositoryModel
from ..models.response import GitHubResponseModel
from ..models.tag import GitHubTagModel
from .base import BaseNamespace
from .contents import GitHubContentsNamespace
from .events import GitHubEventsReposNamespace
from .git import GitHubGitNamespace
from .issues import GitHubIssuesNamespace
from .projects import GitHubRepositoryProjectsNamespace
from .pulls import GitHubPullsNamespace
from .releases import GitHubReleasesNamespace
from .traffic import GitHubTrafficNamespace


class GitHubReposNamespace(BaseNamespace):
    """Methods for the repos namespace"""

    def __post_init__(self) -> None:
        self._contents = GitHubContentsNamespace(self._client)
        self._events = GitHubEventsReposNamespace(self._client)
        self._git = GitHubGitNamespace(self._client)
        self._issues = GitHubIssuesNamespace(self._client)
        self._pulls = GitHubPullsNamespace(self._client)
        self._releases = GitHubReleasesNamespace(self._client)
        self._traffic = GitHubTrafficNamespace(self._client)
        self._projects = GitHubRepositoryProjectsNamespace(self._client)

    @property
    def projects(self) -> GitHubRepositoryProjectsNamespace:
        """Property to access the users projects namespace"""
        return self._projects

    @property
    def contents(self) -> GitHubContentsNamespace:
        """Property to access the contents namespace"""
        return self._contents

    @property
    def events(self) -> GitHubEventsReposNamespace:
        """Property to access the events namespace"""
        return self._events

    @property
    def git(self) -> GitHubGitNamespace:
        """Property to access the git namespace"""
        return self._git

    @property
    def issues(self) -> GitHubIssuesNamespace:
        """Property to access the issues namespace"""
        return self._issues

    @property
    def pulls(self) -> GitHubPullsNamespace:
        """Property to access the pull requests namespace"""
        return self._pulls

    @property
    def releases(self) -> GitHubReleasesNamespace:
        """Property to access the releases namespace"""
        return self._releases

    @property
    def traffic(self) -> GitHubTrafficNamespace:
        """Property to access the traffic namespace"""
        return self._traffic

    async def get(
        self,
        repository: RepositoryType,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubRepositoryModel]:
        """
         Get a repository

         **Arguments**:

         `repository`

         The repository to return, example "octocat/hello-world"

        https://docs.github.com/en/rest/reference/repos#get-a-repository
        """
        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}",
            **kwargs,
        )
        response.data = GitHubRepositoryModel(response.data)
        return response

    async def list_commits(
        self,
        repository: RepositoryType,
        *,
        sha: str | None = None,
        path: str | None = None,
        author: str | None = None,
        since: str | None = None,
        until: str | None = None,
        per_page: int = 30,
        page: int = 0,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[List[GitHubCommitModel]]:
        """
        List commits from a repository.

        **Arguments**:

        - `repository`: The repository to return commits from, specified as
            `<owner>/<repository>`, e.g. "octocat/hello-world".
        - `sha`: Filter the commits to only those with the given SHA.
        - `path`: Filter the commits to only those with changes to the given path.
        - `author`: Filter the commits to only those with the given author.
        - `since`: Filter the commits to only those that were made after the given
            date and time. The `since` parameter should be formatted as an ISO 8601
            date and time, e.g. "2022-12-03T12:34:56Z".
        - `until`: Filter the commits to only those that were made before the given
            date and time. The `until` parameter should be formatted as an ISO 8601
            date and time, e.g. "2022-12-03T12:34:56Z".
        - `per_page`: The number of commits to return per page. The default value is
            30 and the maximum allowed value is 100.
        - `page`: The page number to return. The default value is 0 (the first page).

        For more details, see the API documentation at:
        https://docs.github.com/en/rest/reference/repos#list-commits
        """
        params = {
            "per_page": per_page,
            "page": page,
        }

        if sha is not None:
            params["sha"] = sha
        if path is not None:
            params["path"] = path
        if author is not None:
            params["author"] = author
        if since is not None:
            params["since"] = since
        if until is not None:
            params["until"] = until

        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/commits",
            params=params,
            **kwargs,
        )
        response.data = [GitHubCommitModel(data) for data in response.data]
        return response

    async def list_tags(
        self,
        repository: RepositoryType,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[List[GitHubTagModel]]:
        """
         List tags

         **Arguments**:

         `repository`

         The repository to return tags from, example "octocat/hello-world"

        https://docs.github.com/en/rest/reference/repos#list-repository-tags
        """
        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/tags",
            **kwargs,
        )
        response.data = [GitHubTagModel(data) for data in response.data]
        return response

    async def tarball(
        self,
        repository: RepositoryType,
        *,
        ref: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[bytes]:
        """
         Download a repository archive (tar)

         **Arguments**:

         `repository`

         The repository to return the tar from, example "octocat/hello-world"

         `ref`

         The name of the commit/branch/tag. Default: the repository's default branch (usually main)

        https://docs.github.com/en/rest/reference/repos#download-a-repository-archive-tar
        """

        return await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/tarball/{ref or ''}",
            **kwargs,
        )

    async def zipball(
        self,
        repository: RepositoryType,
        *,
        ref: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[bytes]:
        """
         Download a repository archive (zip)

         **Arguments**:

         `repository`

         The repository to return the zip from, example "octocat/hello-world"

         `ref`

         The name of the commit/branch/tag. Default: the repository's default branch (usually main)

        https://docs.github.com/en/rest/reference/repos#download-a-repository-archive-zip
        """

        return await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/zipball/{ref or ''}",
            **kwargs,
        )

    async def readme(
        self,
        repository: RepositoryType,
        *,
        dir: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[str]:
        """
         Gets the preferred README for a repository.

         **Arguments**:

         `repository`

         The repository to return the zip from, example "octocat/hello-world"

         `dir`

         The alternate path to look for a README file

        https://docs.github.com/en/rest/reference/repos#download-a-repository-archive-zip
        """

        return await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/readme/{dir or ''}",
            **kwargs,
        )



"""GitHub clones data class."""

from __future__ import annotations

from typing import Any, Dict

from .base import GitHubDataModelBase


class _Clone(GitHubDataModelBase):
    """GitHub clone data."""

    timestamp: str | None = None
    count: int | None = None
    uniques: int | None = None


class GitHubClonesModel(GitHubDataModelBase):
    """GitHub clones data class."""

    count: int | None = None
    uniques: int | None = None
    clones: list[_Clone] | None = None

    def _generate_clones(self, data: list[Dict[str, Any]]) -> list[_Clone]:
        """Generate GitHubClonesModel from list of dicts."""
        return [_Clone(clone) for clone in data or []]



"""
Methods for the git namespace

https://docs.github.com/en/rest/reference/git
"""

from __future__ import annotations

from typing import Any, Dict

from ..const import GitHubRequestKwarg, RepositoryType
from ..models.git_tree import GitHubGitTreeModel
from ..models.response import GitHubResponseModel
from .base import BaseNamespace


class GitHubGitNamespace(BaseNamespace):
    """Methods for the git namespace"""

    async def get_tree(
        self,
        repository: RepositoryType,
        tree_sha: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubGitTreeModel]:
        """
        Get a tree
        Returns a single tree using the SHA1 value for that tree.

        If truncated is true in the response then the number of items in
        the tree array exceeded our maximum limit.
        If you need to fetch more items, use the
        non-recursive method of fetching trees, and fetch one sub-tree at a time.

         **Arguments**:

         `repository`

         The repository to return the tree for, example "octocat/hello-world"

         `tree_sha`

         The tree SHA or ref to return

        https://docs.github.com/en/rest/reference/git#get-a-tree
        """
        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository}/git/trees/{tree_sha}",
            **kwargs,
        )
        response.data = GitHubGitTreeModel(response.data)
        return response



"""GitHub milestone data class."""

from __future__ import annotations

from .base import GitHubDataModelBase
from .user import GitHubUserModel


class GitHubMilestoneModel(GitHubDataModelBase):
    """GitHub milestone data class."""

    closed_at: str | None = None
    closed_issues: int | None = None
    created_at: str | None = None
    description: str | None = None
    due_on: str | None = None
    html_url: str | None = None
    id: int | None = None
    labels_url: str | None = None
    number: int | None = None
    open_issues: int | None = None
    state: str | None = None
    title: str | None = None
    updated_at: str | None = None
    url: str | None = None
    creator: GitHubUserModel | None = None

    def _generate_creator(self, data: dict) -> GitHubUserModel:
        """Generate the creator."""
        return GitHubUserModel(data)



"""GitHub response data class."""

from __future__ import annotations

from typing import Any, Dict

from .base import GitHubDataModelBase
from .license import GitHubLicenseModel
from .organization import GitHubOrganizationModel
from .owner import GitHubOwnerModel
from .permissions import GitHubPermissionsModel


class GitHubRepositoryModel(GitHubDataModelBase):
    """GitHub repository data class."""

    archive_url: str | None = None
    allow_squash_merge: bool | None = None
    allow_merge_commit: bool | None = None
    allow_rebase_merge: bool | None = None
    allow_auto_merge: bool | None = None
    delete_branch_on_merge: bool | None = None
    archived: bool | None = None
    assignees_url: str | None = None
    blobs_url: str | None = None
    branches_url: str | None = None
    clone_url: str | None = None
    collaborators_url: str | None = None
    comments_url: str | None = None
    commits_url: str | None = None
    compare_url: str | None = None
    contents_url: str | None = None
    contributors_url: str | None = None
    created_at: str | None = None
    default_branch: str | None = None
    deployments_url: str | None = None
    description: str | None = None
    disabled: bool | None = None
    downloads_url: str | None = None
    events_url: str | None = None
    fork: bool | None = None
    forks_count: int | None = None
    forks_url: str | None = None
    forks: int | None = None
    full_name: str | None = None
    git_commits_url: str | None = None
    git_refs_url: str | None = None
    git_tags_url: str | None = None
    git_url: str | None = None
    has_downloads: bool | None = None
    has_issues: bool | None = None
    has_pages: bool | None = None
    has_projects: bool | None = None
    has_wiki: bool | None = None
    homepage: str | None = None
    hooks_url: str | None = None
    html_url: str | None = None
    id: int | None = None
    issue_comment_url: str | None = None
    issue_events_url: str | None = None
    issues_url: str | None = None
    keys_url: str | None = None
    labels_url: str | None = None
    language: str | None = None
    languages_url: str | None = None
    license: GitHubLicenseModel | None = None
    merges_url: str | None = None
    milestones_url: str | None = None
    mirror_url: str | None = None
    name: str | None = None
    network_count: int | None = None
    node_id: str | None = None
    notifications_url: str | None = None
    open_issues_count: int | None = None
    open_issues: int | None = None
    organization: GitHubOrganizationModel | None = None
    owner: GitHubOwnerModel | None = None
    permissions: GitHubPermissionsModel | None = None
    private: bool | None = None
    pulls_url: str | None = None
    pushed_at: str | None = None
    releases_url: str | None = None
    size: int | None = None
    ssh_url: str | None = None
    stargazers_count: int | None = None
    stargazers_url: str | None = None
    statuses_url: str | None = None
    subscribers_count: int | None = None
    subscribers_url: str | None = None
    subscription_url: str | None = None
    svn_url: str | None = None
    tags_url: str | None = None
    teams_url: str | None = None
    temp_clone_token: str | None = None
    topics: list[str] | None = None
    trees_url: str | None = None
    updated_at: str | None = None
    url: str | None = None
    watchers_count: int | None = None
    watchers: int | None = None

    def _generate_license(self, data: Dict[str, Any] | None) -> GitHubLicenseModel:
        """Generate a license model from a dictionary."""
        return GitHubLicenseModel(data) if data else None

    def _generate_owner(self, data: Dict[str, Any] | None) -> GitHubOwnerModel:
        """Generate an owner model from a dictionary."""
        return GitHubOwnerModel(data) if data else None

    def _generate_permissions(self, data: Dict[str, Any] | None) -> GitHubPermissionsModel:
        """Generate a permissions model from a dictionary."""
        return GitHubPermissionsModel(data)

    def _generate_organization(self, data: Dict[str, Any] | None) -> GitHubOrganizationModel:
        """Generate an organization model from a dictionary."""
        return GitHubOrganizationModel(data)



"""This is the class that do the requests against the API."""

from __future__ import annotations

import asyncio
from typing import Any, Dict

import aiohttp

from .const import (
    GitHubClientKwarg,
    GitHubRequestKwarg,
    HttpContentType,
    HttpMethod,
    HttpStatusCode,
)
from .exceptions import (
    GitHubAuthenticationException,
    GitHubConnectionException,
    GitHubException,
    GitHubGraphQLException,
    GitHubNotFoundException,
    GitHubNotModifiedException,
    GitHubPayloadException,
    GitHubPermissionException,
    GitHubRatelimitException,
)
from .legacy.client import AIOGitHubAPIClient as LegacyAIOGitHubAPIClient
from .models.base import GitHubBase
from .models.request_data import GitHubBaseRequestDataModel
from .models.response import GitHubResponseModel

STATUS_EXCEPTIONS: Dict[HttpStatusCode, GitHubException] = {
    HttpStatusCode.FORBIDDEN: GitHubAuthenticationException,
    HttpStatusCode.UNAUTHORIZED: GitHubAuthenticationException,
    HttpStatusCode.NOT_MODIFIED: GitHubNotModifiedException,
    HttpStatusCode.NOT_FOUND: GitHubNotFoundException,
    HttpStatusCode.BAD_REQUEST: GitHubPayloadException,
    HttpStatusCode.UNPROCESSABLE_ENTITY: GitHubException,
}

MESSAGE_EXCEPTIONS: Dict[str, GitHubException] = {
    "Bad credentials": GitHubAuthenticationException,
    "You have exceeded a secondary rate limit and have been temporarily blocked from content creation. Please retry your request again later.": GitHubRatelimitException,
    "Must have push access to repository": GitHubPermissionException,
}


class AIOGitHubAPIClient(LegacyAIOGitHubAPIClient):
    """Dummy class to not break existing code."""


class GitHubClient(GitHubBase):
    """
    Client to handle API calls.

    Don't use this directly, use `aiogithubapi.github.GitHubApi` to get the client.
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        token: str | None = None,
        *,
        api_version: str | None = None,
        **kwargs: Dict[GitHubClientKwarg, Any],
    ) -> None:
        """
        Initialise the GitHub API client.

        **Arguments:**
        `session`: The aiohttp client session to use for making requests.
        `token`: The GitHub access token to use for authenticating requests. Can be a string or None.
        `api_version`: The version of the GitHub API to use. Can be a string or None.

        """
        self._base_request_data = GitHubBaseRequestDataModel(
            token=token,
            api_version=api_version,
            kwargs=kwargs,
        )
        self._session = session
        self._loop = asyncio.get_running_loop()

    async def async_call_api(
        self,
        endpoint: str,
        *,
        data: Dict[str, Any] | str | None = None,
        headers: Dict[str, Any] | None = None,
        method: HttpMethod = HttpMethod.GET,
        params: Dict[str, Any] | None = None,
        timeout: int | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel:
        """
        Makes an HTTP request to the specified endpoint using the specified parameters.

        This method is asynchronous, meaning that it will not block the execution of the program while the request is being made and processed.

        **Arguments**:

        -  `endpoint` (Required): The API endpoint to call.

        **Optional arguments**:
        - `endpoint`: The API endpoint to call.
        - `data`: The data to include in the request body. Can be a dictionary, a string, or None.
        - `headers`: The headers to include in the request. Can be a dictionary or None.
        - `method`: The HTTP method to use for the request. Defaults to GET.
        - `params`: The query parameters to include in the request. Can be a dictionary or None.
        - `timeout`: The maximum amount of time to wait for the request to complete, in seconds. Can be an integer or None.

        Returns:
        A GitHubResponseModel object representing the API response.
        """
        request_arguments: Dict[str, Any] = {
            "url": self._base_request_data.request_url(endpoint),
            "method": kwargs.get(GitHubRequestKwarg.METHOD, method).lower(),
            "params": params
            or kwargs.get(GitHubRequestKwarg.PARAMS, kwargs.get(GitHubRequestKwarg.QUERY, {})),
            "timeout": timeout or self._base_request_data.timeout,
            "headers": {
                **(headers or {}),
                **self._base_request_data.headers,
                **(headers or {}),
                **kwargs.get("headers", {}),
            },
        }

        if etag := kwargs.get(GitHubRequestKwarg.ETAG):
            request_arguments["headers"][aiohttp.hdrs.IF_NONE_MATCH] = etag

        if isinstance(data, dict):
            request_arguments["json"] = data
        else:
            request_arguments["data"] = data

        try:
            result = await self._session.request(**request_arguments)
        except (aiohttp.ClientError, asyncio.CancelledError) as exception:
            raise GitHubConnectionException(
                "Request exception for "
                f"'{self._base_request_data.request_url(endpoint)}' with - {exception}"
            ) from exception

        except asyncio.TimeoutError:
            raise GitHubConnectionException(
                f"Timeout of {self._base_request_data.timeout} reached while "
                f"waiting for {self._base_request_data.request_url(endpoint)}"
            ) from None

        except BaseException as exception:
            raise GitHubException(
                "Unexpected exception for "
                f"'{self._base_request_data.request_url(endpoint)}' with - {exception}"
            ) from exception

        response = GitHubResponseModel(result)
        if response.status == HttpStatusCode.NO_CONTENT:
            return response

        try:
            if HttpContentType.BASE_JSON in (response.headers.content_type or ""):
                response.data = await result.json(encoding="utf-8")
            elif (response.headers.content_type or "") in (
                HttpContentType.BASE_ZIP,
                HttpContentType.BASE_GZIP,
            ):
                response.data = await result.read()
            else:
                response.data = await result.text(encoding="utf-8")
        except BaseException as exception:
            raise GitHubException(
                f"Could not handle response data from '{self._base_request_data.request_url(endpoint)}' with - {exception}"
            )
        message = response.data.get("message") if isinstance(response.data, dict) else None

        if message is not None and "rate limit" in message:
            raise GitHubRatelimitException(message)

        if exception := STATUS_EXCEPTIONS.get(response.status):
            raise exception(message or response.data)

        if isinstance(response.data, dict):
            if message is not None:
                if exception := MESSAGE_EXCEPTIONS.get(message):
                    raise exception(message)
                # For a 201 Created response, we assume the operation was successful and ignore any "message" field.
                if response.status != HttpStatusCode.CREATED:
                    raise GitHubException(message)

        if endpoint == "/graphql" and response.data.get("errors", []):
            raise GitHubGraphQLException(
                ", ".join(entry.get("message") for entry in response.data["errors"])
            )

        return response



"""Fixtures."""
import pytest


@pytest.fixture()
def base_response():
    return {"key": "value"}


@pytest.fixture()
def bad_auth_response():
    return {"message": "Bad credentials"}


@pytest.fixture()
def bad_response():
    return {"message": "I'm a tea pot", "code": 418}



"""
Generated by generate/generate.py - 2020-09-08 17:49:51.525706
"""
import pytest


@pytest.fixture()
def device_fixtrue_response():
    return {
        "device_code": "3584d83530557fdd1f46af8289938c8ef79f9dc5",
        "user_code": "WDJB-MJHT",
        "verification_uri": "https://github.com/login/device",
        "expires_in": 900,
        "interval": 5,
    }



"""GitHub pull_request data model."""

from __future__ import annotations

from .issue import GitHubIssueModel


class GitHubPullRequestModel(GitHubIssueModel):
    """GitHub pull_request data model."""

    diff_url: str | None = None
    patch_url: str | None = None
    issue_url: str | None = None
    merged_at: str | None = None
    merge_commit_sha: str | None = None
    requested_reviewers: list[dict] | None = None
    requested_teams: list[dict] | None = None
    draft: bool | None = None
    head: dict | None = None
    base: dict | None = None
    auto_merge: bool | None = None
    commits_url: str | None = None
    review_comments_url: str | None = None
    review_comment_url: str | None = None
    statuses_url: str | None = None



from datetime import datetime
from unittest.mock import AsyncMock

import aiohttp
import pytest

from aiogithubapi import AIOGitHubAPIException, GitHubDevice, GitHubDeviceAPI

from tests.common import CLIENT_ID, load_fixture
from tests.conftest import client_session


@pytest.mark.asyncio
async def test_session():
    async with aiohttp.ClientSession() as session:
        device = GitHubDevice("xxxxx", "", session)
        assert device.session == session

    async with GitHubDevice("xxxxx") as github:
        assert github.session is not None


@pytest.mark.asyncio
async def test_device(mock_response, github_device: GitHubDevice, asyncio_sleep: AsyncMock):
    device = await github_device.async_register_device()
    assert device.user_code == "WDJB-MJHT"

    mock_response.mock_data_list = [
        {
            "error": "authorization_pending",
            "error_description": "Pending user authorization",
        },
        {
            "access_token": "e72e16c7e42f292c6912e7710c838347ae178b4a",
            "token_type": "bearer",
            "scope": "user",
        },
    ]

    activation = await github_device.async_device_activation()
    assert asyncio_sleep.call_count == 1
    assert asyncio_sleep.call_args[0][0] == 1
    assert activation.access_token == "e72e16c7e42f292c6912e7710c838347ae178b4a"


@pytest.mark.asyncio
async def test_no_user_response(github_device: GitHubDevice):
    device = await github_device.async_register_device()
    assert device.user_code == "WDJB-MJHT"

    with pytest.raises(AIOGitHubAPIException, match="User took too long to enter key"):
        github_device._expires = datetime.timestamp(datetime.now())
        await github_device.async_device_activation()


@pytest.mark.asyncio
async def test_device_error(mock_response, github_device: GitHubDevice):
    device = await github_device.async_register_device()
    assert device.user_code == "WDJB-MJHT"

    mock_response.mock_data = load_fixture("oauth_access_token_error.json", True)

    with pytest.raises(AIOGitHubAPIException, match="Unsupported grant type"):
        await github_device.async_device_activation()


@pytest.mark.asyncio
async def test_no_device_code(github_device: GitHubDevice, asyncio_sleep: AsyncMock):
    with pytest.raises(AIOGitHubAPIException, match="User took too long to enter key"):
        github_device._expires = datetime.timestamp(datetime.now())
        await github_device.async_device_activation()
        assert asyncio_sleep.call_count == 1
        assert asyncio_sleep.call_args[0][0] == 5



"""Base class for all GitHub objects."""

from __future__ import annotations

from logging import Logger
from typing import Any, Dict

from ..const import LOGGER


class GitHubBase:
    """Base class for all GitHub objects."""

    logger: Logger = LOGGER

    @staticmethod
    def slugify(value: str) -> str:
        """Slugify."""
        return str(value).replace("-", "_").lower()


class GitHubDataModelBase(GitHubBase):
    """Base class for all GitHub data objects."""

    _raw_data: Any | None = None
    _log_missing: bool = True
    _process_data: bool = True
    _slugify_keys: bool = True

    def __init__(self, data: Dict[str, Any]) -> None:
        """Init."""
        self._raw_data = data
        if self._process_data:
            for key, value in self._raw_data.items():
                if self._slugify_keys:
                    key = self.slugify(key)
                if hasattr(self, key):
                    if handler := getattr(self, f"_generate_{key}", None):
                        value = handler(value)
                    self.__setattr__(key, value)
        self.__post_init__()

    def __post_init__(self):
        """Post init."""

    @property
    def as_dict(self) -> Dict[str, Any]:
        """Return attributes as a dict."""

        def expand_value_if_needed(value: Any) -> Any:
            if isinstance(value, GitHubDataModelBase):
                return value.as_dict
            if isinstance(value, list):
                return [expand_value_if_needed(v) for v in value]
            return value

        return {
            key: expand_value_if_needed(value)
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        }



"""
AIOGitHubAPI: Issue

https://developer.github.com/v3/issues/
"""

# pylint: disable=missing-docstring, unused-import
from ...base import AIOGitHubAPIBaseClient
from ...repository.issue.comment import (
    AIOGitHubAPIRepositoryIssueComment,
    AIOGitHubAPIRepositoryIssueCommentUser,
)


class AIOGitHubAPIRepositoryIssue(AIOGitHubAPIBaseClient):
    """Issue comment user GitHub API implementation."""

    @property
    def html_url(self):
        return self.attributes.get("html_url")

    @property
    def id(self):
        return self.attributes.get("id")

    @property
    def number(self):
        return self.attributes.get("number")

    @property
    def labels(self):
        return self.attributes.get("labels", [])

    @property
    def title(self):
        return self.attributes.get("title")

    @property
    def state(self):
        return self.attributes.get("state")

    @property
    def assignees(self):
        return self.attributes.get("assignees")

    @property
    def body(self):
        return self.attributes.get("body")

    @property
    def repository(self):
        repository_url = self.attributes.get("repository_url").split("/")
        return f"{repository_url[-2]}/{repository_url[-1]}"

    @property
    def user(self):
        return AIOGitHubAPIRepositoryIssueCommentUser(self.attributes.get("user"))

    async def get_comments(
        self,
    ) -> ["AIOGitHubAPIRepositoryIssueComment"] or list:
        """Updates an issue comment."""
        _endpoint = f"/repos/{self.repository}/issues/{self.id}/comments"

        response = await self.client.get(endpoint=_endpoint)

        return [
            AIOGitHubAPIRepositoryIssueComment(self.client, x, self.repository)
            for x in response.data or []
        ]

    async def comment(self, body: str) -> None:
        """Adds a comment to an issue."""
        _endpoint = f"/repos/{self.repository}/issues/{self.id}/comments"

        await self.client.post(endpoint=_endpoint, data={"body": body}, jsondata=True)

    async def update(
        self,
        title: str or None = None,
        body: str or None = None,
        state: str or None = None,
        milestone: int or None = None,
        labels: [str] or None = None,
        assignees: [str] or None = None,
    ):
        """Updates an issue comment."""
        _endpoint = f"/repos/{self.repository}/issues/{self.id}"

        data = {}
        if title is not None:
            data["title"] = title
        if body is not None:
            data["body"] = body
        if state is not None:
            data["state"] = state
        if milestone is not None:
            data["milestone"] = milestone
        if labels is not None:
            data["labels"] = labels
        if assignees is not None:
            data["assignees"] = assignees

        await self.client.post(endpoint=_endpoint, data=data, jsondata=True)



"""
Generated by generate/generate.py - 2020-08-02 12:24:23.865974
"""
from aiogithubapi.objects.repos.traffic.clones import AIOGitHubAPIReposTrafficClones

from tests.legacy.responses.repos.traffic.clones_fixtrue import clones_fixtrue_response


def test_clones(clones_fixtrue_response):
    obj = AIOGitHubAPIReposTrafficClones(clones_fixtrue_response)
    assert obj.count == clones_fixtrue_response["count"]
    assert obj.uniques == clones_fixtrue_response["uniques"]
    assert obj.clones[0].timestamp == clones_fixtrue_response["clones"][0]["timestamp"]
    assert obj.clones[0].count == clones_fixtrue_response["clones"][0]["count"]
    assert obj.clones[0].uniques == clones_fixtrue_response["clones"][0]["uniques"]



"""Test device_flow."""
# pylint: disable=missing-docstring,protected-access
from datetime import datetime
from unittest.mock import AsyncMock
from aiogithubapi.exceptions import GitHubAuthenticationException

import pytest

from aiogithubapi import (
    DeviceFlowError,
    GitHubDeviceAPI,
    GitHubException,
    GitHubRequestKwarg,
)

from tests.common import CLIENT_ID, DEVICE_CODE, TOKEN, MockedRequests, MockResponse


@pytest.mark.asyncio
async def test_register_device(
    github_device_api: GitHubDeviceAPI,
    mock_requests: MockedRequests,
):
    response = await github_device_api.register()
    assert response.status == 200
    assert response.data.verification_uri == "https://github.com/login/device"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://github.com/login/device/code"
    assert mock_requests.last_request["params"]["scope"] == ""
    assert mock_requests.last_request["params"]["client_id"] == CLIENT_ID


@pytest.mark.asyncio
async def test_register_device_custom_scope(
    github_device_api: GitHubDeviceAPI,
    mock_requests: MockedRequests,
):
    response = await github_device_api.register(**{GitHubRequestKwarg.SCOPE: "custom"})
    assert response.status == 200
    assert response.data.verification_uri == "https://github.com/login/device"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://github.com/login/device/code"
    assert mock_requests.last_request["params"]["scope"] == "custom"


@pytest.mark.asyncio
async def test_device_activation(
    github_device_api: GitHubDeviceAPI,
    mock_requests: MockedRequests,
):
    response = await github_device_api.activation(device_code=DEVICE_CODE)
    assert response.status == 200
    assert response.data.access_token == TOKEN
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://github.com/login/oauth/access_token"


@pytest.mark.asyncio
async def test_missing_expiration(github_device_api: GitHubDeviceAPI):
    github_device_api._expires = None
    with pytest.raises(GitHubException, match="Expiration has passed, re-run the registration"):
        await github_device_api.activation(device_code=DEVICE_CODE)


@pytest.mark.asyncio
async def test_user_too_slow(github_device_api: GitHubDeviceAPI):
    github_device_api._expires = datetime.timestamp(datetime.now()) - 1
    with pytest.raises(GitHubException, match="User took too long to enter key"):
        await github_device_api.activation(device_code=DEVICE_CODE)


@pytest.mark.asyncio
async def test_wait_for_confirmation(
    github_device_api: GitHubDeviceAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
    caplog: pytest.CaptureFixture,
    asyncio_sleep: AsyncMock,
):
    mock_response.mock_data_list = [
        {
            "error": DeviceFlowError.AUTHORIZATION_PENDING,
            "error_description": "Pending user authorization",
        },
        {
            "error": DeviceFlowError.AUTHORIZATION_PENDING,
            "error_description": "Pending user authorization",
        },
        {
            "error": DeviceFlowError.SLOW_DOWN,
            "error_description": "Too many requests have been made in the same timeframe.",
            "interval": 20,
        },
        {
            "error": DeviceFlowError.AUTHORIZATION_PENDING,
            "error_description": "Pending user authorization",
        },
        {
            "access_token": "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
            "token_type": "bearer",
            "scope": "user",
        },
    ]
    response = await github_device_api.activation(device_code=DEVICE_CODE)
    assert response.status == 200
    assert mock_requests.called == 5
    assert asyncio_sleep.call_count == 4

    # Use default interval
    assert asyncio_sleep.call_args_list[-3][0][0] == 1
    # Use new interval from API for all next calls
    assert asyncio_sleep.call_args_list[-2][0][0] == 20
    assert asyncio_sleep.call_args_list[-1][0][0] == 20

    assert mock_requests.last_request["url"] == "https://github.com/login/oauth/access_token"
    assert "Pending user authorization" in caplog.text
    assert "Got new interval instruction of 20 from the API" in caplog.text


@pytest.mark.asyncio
async def test_error_while_waiting(
    github_device_api: GitHubDeviceAPI,
    mock_response: MockResponse,
):
    mock_response.mock_data = {"error": "any", "error_description": "Any error message"}
    with pytest.raises(GitHubException, match="Any error message"):
        await github_device_api.activation(device_code=DEVICE_CODE)


@pytest.mark.asyncio
async def test_secondary_rate_limit(
    github_device_api: GitHubDeviceAPI,
    mock_response: MockResponse,
):
    mock_response.mock_data = "<html><p>You have exceeded a secondary rate limit.<br></html>"
    with pytest.raises(GitHubAuthenticationException, match="Secondary rate limit exceeded"):
        await github_device_api.activation(device_code=DEVICE_CODE)


"""
Class object for AIOGitHubAPIRepositoryCollaborator
Documentation: https://docs.github.com/en/rest/reference/repos#list-repository-collaborators

Generated by generate/generate.py - 2020-08-02 10:09:26.842173
"""

from ..base import AIOGitHubAPIBase


class Permissions(AIOGitHubAPIBase):
    @property
    def pull(self):
        return self.attributes.get("pull", True)

    @property
    def push(self):
        return self.attributes.get("push", True)

    @property
    def admin(self):
        return self.attributes.get("admin", False)


class AIOGitHubAPIRepositoryCollaborator(AIOGitHubAPIBase):
    @property
    def login(self):
        return self.attributes.get("login", "")

    @property
    def id(self):
        return self.attributes.get("id", None)

    @property
    def node_id(self):
        return self.attributes.get("node_id", "")

    @property
    def avatar_url(self):
        return self.attributes.get("avatar_url", "")

    @property
    def gravatar_id(self):
        return self.attributes.get("gravatar_id", "")

    @property
    def url(self):
        return self.attributes.get("url", "")

    @property
    def html_url(self):
        return self.attributes.get("html_url", "")

    @property
    def followers_url(self):
        return self.attributes.get("followers_url", "")

    @property
    def following_url(self):
        return self.attributes.get("following_url", "")

    @property
    def gists_url(self):
        return self.attributes.get("gists_url", "")

    @property
    def starred_url(self):
        return self.attributes.get("starred_url", "")

    @property
    def subscriptions_url(self):
        return self.attributes.get("subscriptions_url", "")

    @property
    def organizations_url(self):
        return self.attributes.get("organizations_url", "")

    @property
    def repos_url(self):
        return self.attributes.get("repos_url", "")

    @property
    def events_url(self):
        return self.attributes.get("events_url", "")

    @property
    def received_events_url(self):
        return self.attributes.get("received_events_url", "")

    @property
    def type(self):
        return self.attributes.get("type", "")

    @property
    def site_admin(self):
        return self.attributes.get("site_admin", False)

    @property
    def permissions(self):
        return Permissions(self.attributes.get("permissions", {}))



"""Test projects namespace."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI, GitHubProjectModel

from tests.common import (
    TEST_ORGANIZATION,
    TEST_REPOSITORY_NAME,
    MockedRequests,
    MockResponse,
    load_fixture,
)


@pytest.mark.asyncio
async def test_list_organization_projects(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.orgs.projects.list(TEST_ORGANIZATION)
    assert response.status == 200
    assert isinstance(response.data, list)
    fist = response.data[0]
    assert isinstance(fist, GitHubProjectModel)
    assert fist.name == "Organization Roadmap"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/orgs/octocat/projects"


@pytest.mark.asyncio
async def test_create_organization_project(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "orgs_octocat_projects.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.orgs.projects.create(TEST_ORGANIZATION, name="test", body="test")
    assert response.status == 200
    assert isinstance(response.data, GitHubProjectModel)
    assert response.data.name == "Organization Roadmap"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/orgs/octocat/projects"


@pytest.mark.asyncio
async def test_list_repository_projects(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.projects.list(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    fist = response.data[0]
    assert isinstance(fist, GitHubProjectModel)
    assert fist.name == "Projects Documentation"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/projects"
    )


@pytest.mark.asyncio
async def test_create_repository_project(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_projects.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.projects.create(
        TEST_REPOSITORY_NAME, name="test", body="test"
    )
    assert response.status == 200
    assert isinstance(response.data, GitHubProjectModel)
    assert response.data.name == "Projects Documentation"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/projects"
    )


@pytest.mark.asyncio
async def test_list_users_projects(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.users.projects.list(TEST_ORGANIZATION)
    assert response.status == 200
    assert isinstance(response.data, list)
    fist = response.data[0]
    assert isinstance(fist, GitHubProjectModel)
    assert fist.name == "My Projects"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/users/octocat/projects"


@pytest.mark.asyncio
async def test_create_user_project(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "users_octocat_projects.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.user.projects.create(name="test", body="test")
    assert response.status == 200
    assert isinstance(response.data, GitHubProjectModel)
    assert response.data.name == "My Projects"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/user/projects"


@pytest.mark.asyncio
async def test_base_get_project(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.projects.get(project_id=1)
    assert response.status == 200
    assert isinstance(response.data, GitHubProjectModel)
    assert response.data.name == "Projects Documentation"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/projects/1"


@pytest.mark.asyncio
async def test_base_update_project(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.projects.update(project_id=1, data={"test": "update"})
    assert response.status == 200
    assert isinstance(response.data, GitHubProjectModel)
    assert response.data.name == "Projects Documentation"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/projects/1"


@pytest.mark.asyncio
async def test_base_delete_project(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.projects.delete(project_id=1)
    assert response.status == 200
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/projects/1"



"""
Class object for AIOGitHubAPILoginOauth
Documentation: https://docs.github.com/en/developers/apps/authorizing-oauth-apps#device-flow

Generated by generate/generate.py - 2020-09-08 18:45:20.581868
"""

from ..base import AIOGitHubAPIBase


class AIOGitHubAPILoginOauth(AIOGitHubAPIBase):
    @property
    def access_token(self):
        return self.attributes.get("access_token", "")

    @property
    def token_type(self):
        return self.attributes.get("token_type", "")

    @property
    def scope(self):
        return self.attributes.get("scope", "")



"""
Generated by generate/generate.py - 2020-08-02 10:33:09.248066
"""
import pytest


@pytest.fixture()
def team_fixtrue_response():
    return {
        "id": 1,
        "node_id": "MDQ6VGVhbTE=",
        "url": "https://api.github.com/teams/1",
        "html_url": "https://api.github.com/teams/justice-league",
        "name": "Justice League",
        "slug": "justice-league",
        "description": "A great team.",
        "privacy": "closed",
        "permission": "admin",
        "members_url": "https://api.github.com/teams/1/members{/member}",
        "repositories_url": "https://api.github.com/teams/1/repos",
        "parent": None,
        "members_count": 3,
        "repos_count": 10,
        "created_at": "2017-07-14T16:53:42Z",
        "updated_at": "2017-08-17T12:37:15Z",
        "organization": {
            "login": "github",
            "id": 1,
            "node_id": "MDEyOk9yZ2FuaXphdGlvbjE=",
            "url": "https://api.github.com/orgs/github",
            "repos_url": "https://api.github.com/orgs/github/repos",
            "events_url": "https://api.github.com/orgs/github/events",
            "hooks_url": "https://api.github.com/orgs/github/hooks",
            "issues_url": "https://api.github.com/orgs/github/issues",
            "members_url": "https://api.github.com/orgs/github/members{/member}",
            "public_members_url": "https://api.github.com/orgs/github/public_members{/member}",
            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
            "description": "A great organization",
            "name": "github",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "octocat@github.com",
            "twitter_username": "github",
            "is_verified": True,
            "has_organization_projects": True,
            "has_repository_projects": True,
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/octocat",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "Organization",
        },
    }



"""Used for the GitHub API namespace."""

from ..client import GitHubClient


class BaseNamespace:
    """Used for the GitHub API namespace."""

    def __init__(self, client: GitHubClient) -> None:
        """Initialise the namespace."""
        self._client = client
        self.__post_init__()

    def __post_init__(self) -> None:
        """Post initialisation."""



"""GitHub label data class."""

from __future__ import annotations

from .base import GitHubDataModelBase


class GitHubLabelModel(GitHubDataModelBase):
    """GitHub label data class."""

    color: str | None = None
    default: bool | None = None
    description: str | None = None
    id: int | None = None
    name: str | None = None
    url: str | None = None



"""
Generated by generate/generate.py - 2020-08-02 10:09:26.842173
"""
from aiogithubapi.objects.repository.collaborator import (
    AIOGitHubAPIRepositoryCollaborator,
)

from tests.legacy.responses.repository.collaborator_fixtrue import (
    collaborator_fixtrue_response,
)


def test_collaborator(collaborator_fixtrue_response):
    obj = AIOGitHubAPIRepositoryCollaborator(collaborator_fixtrue_response)
    assert obj.login == collaborator_fixtrue_response["login"]
    assert obj.id == collaborator_fixtrue_response["id"]
    assert obj.node_id == collaborator_fixtrue_response["node_id"]
    assert obj.avatar_url == collaborator_fixtrue_response["avatar_url"]
    assert obj.gravatar_id == collaborator_fixtrue_response["gravatar_id"]
    assert obj.url == collaborator_fixtrue_response["url"]
    assert obj.html_url == collaborator_fixtrue_response["html_url"]
    assert obj.followers_url == collaborator_fixtrue_response["followers_url"]
    assert obj.following_url == collaborator_fixtrue_response["following_url"]
    assert obj.gists_url == collaborator_fixtrue_response["gists_url"]
    assert obj.starred_url == collaborator_fixtrue_response["starred_url"]
    assert obj.subscriptions_url == collaborator_fixtrue_response["subscriptions_url"]
    assert obj.organizations_url == collaborator_fixtrue_response["organizations_url"]
    assert obj.repos_url == collaborator_fixtrue_response["repos_url"]
    assert obj.events_url == collaborator_fixtrue_response["events_url"]
    assert obj.received_events_url == collaborator_fixtrue_response["received_events_url"]
    assert obj.type == collaborator_fixtrue_response["type"]
    assert obj.site_admin == collaborator_fixtrue_response["site_admin"]
    assert obj.permissions.pull == collaborator_fixtrue_response["permissions"]["pull"]
    assert obj.permissions.push == collaborator_fixtrue_response["permissions"]["push"]
    assert obj.permissions.admin == collaborator_fixtrue_response["permissions"]["admin"]



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import json

import pytest

from aiogithubapi import GitHub
from aiogithubapi.common.exceptions import (
    AIOGitHubAPINotModifiedException,
    AIOGitHubAPIRatelimitException,
)

from tests.common import TOKEN
from tests.legacy.responses.base import base_response
from tests.legacy.responses.org_repositories import org_repositories_response


@pytest.mark.asyncio
async def test_get_org_repos(mock_response, org_repositories_response, client_session):
    mock_response.mock_data = org_repositories_response

    async with GitHub(TOKEN, session=client_session) as github:
        org = await github.get_org_repos("octocat")
        first_repo = org[0]
        assert first_repo.description == "This your first repo!"
        assert github.client.ratelimits.remaining == "4999"

        mock_response.clear()
        mock_response.mock_status = 304

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await github.get_org_repos("octocat", etag=github.client.last_response.etag)


@pytest.mark.asyncio
async def test_get_org_repos_ratelimited(github, mock_response):
    mock_response.mock_status = 403

    with pytest.raises(AIOGitHubAPIRatelimitException):
        await github.get_org_repos("octocat")



"""
AIOGitHubAPI: AioGitHubClient

This is the class that do the requests against the API
It also keeps track of ratelimits
"""

# pylint: disable=redefined-builtin, too-many-arguments
from __future__ import annotations

from typing import Optional

import aiohttp

from ..common.const import BASE_API_HEADERS, BASE_API_URL
from ..helpers import async_call_api
from ..objects.base import AIOGitHubAPIBase, AIOGitHubAPIResponse
from ..objects.ratelimit import AIOGitHubAPIRateLimit


class AIOGitHubAPIClient(AIOGitHubAPIBase):
    """
    Client to handle API calls.

    Deprecated: Used by old versions of the library.
    """

    def __init__(
        self,
        session: aiohttp.ClientSession,
        token: str,
        headers: Optional[dict] = None,
        base_url: Optional[str] = None,
    ) -> None:
        """Initialize the API client."""
        self.session = session
        self.base_url = base_url or BASE_API_URL
        self.last_response: Optional[AIOGitHubAPIResponse] = None
        self.token = token
        self.ratelimits = AIOGitHubAPIRateLimit()
        self.headers = {}
        self.headers.update(BASE_API_HEADERS)
        self.headers.update(headers or {})

        if token is not None:
            self.headers["Authorization"] = "token {}".format(token)

    async def get(
        self,
        endpoint: str,
        returnjson: bool = True,
        headers: dict or None = None,
        params: dict or None = None,
    ) -> AIOGitHubAPIResponse:
        """Execute a GET request."""
        url = f"{self.base_url}{endpoint}"
        _headers = {}
        _headers.update(self.headers)
        _headers.update(headers or {})

        response = await async_call_api(
            session=self.session,
            method="GET",
            url=url,
            returnjson=returnjson,
            headers=_headers,
            params=params,
        )
        self.ratelimits.load_from_response_headers(response.headers)
        self.last_response = response
        return response

    async def post(
        self,
        endpoint: str,
        returnjson: bool = False,
        headers: dict or None = None,
        params: dict or None = None,
        data: dict or str or None = None,
        jsondata: bool = False,
    ) -> AIOGitHubAPIResponse:
        """Execute a POST request."""
        url = f"{self.base_url}{endpoint}"
        _headers = {}
        _headers.update(self.headers)
        _headers.update(headers or {})

        response = await async_call_api(
            session=self.session,
            method="POST",
            url=url,
            returnjson=returnjson,
            headers=_headers,
            params=params,
            data=data,
            jsondata=jsondata,
        )
        self.ratelimits.load_from_response_headers(response.headers)
        self.last_response = response
        return response



"""
Generated by generate/generate.py - 2020-08-02 10:32:07.005845
"""
import pytest


@pytest.fixture()
def label_fixtrue_response():
    return {
        "id": 208045946,
        "node_id": "MDU6TGFiZWwyMDgwNDU5NDY=",
        "url": "https://api.github.com/repos/octocat/Hello-World/labels/bug",
        "name": "bug",
        "description": "Something isn't working",
        "color": "f29513",
        "default": True,
    }



"""GitHub Activity event models"""

from __future__ import annotations

from typing import Any, Dict

from .base import GitHubDataModelBase


class _Actor(GitHubDataModelBase):
    """Internal entry."""

    id: str | None = None
    login: str | None = None
    display_login: str | None = None
    gravatar_id: str | None = None
    url: str | None = None
    avatar_url: str | None = None


class _Repo(GitHubDataModelBase):
    """Internal entry."""

    id: str | None = None
    name: str | None = None
    url: str | None = None


class GitHubEventModel(GitHubDataModelBase):
    """
    GitHub base event data class.

    The type is PascalCase

    https://docs.github.com/en/developers/webhooks-and-events/events/github-event-types
    """

    id: str | None = None
    type: str | None = None
    actor: _Actor | None = None
    org: _Actor | None = None
    repo: _Repo | None = None
    payload: Dict[str, Any] | None = None
    public: bool | None = None
    created_at: str | None = None

    def _generate_actor(self, data: Dict[str, Any] | None) -> _Actor:
        """Generate actor data."""
        return _Actor(data) if data else None

    def _generate_repo(self, data: Dict[str, Any] | None) -> _Repo:
        """Generate repo data."""
        return _Repo(data) if data else None



"""Test graphql."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI, GitHubGraphQLException
from aiogithubapi.graphql_examples.repository_tags import EXAMPLE_QUERY

from tests.common import MockedRequests, MockResponse


@pytest.mark.asyncio
async def test_graphql(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.graphql(EXAMPLE_QUERY)
    assert response.status == 200
    assert isinstance(response.data, dict)
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/graphql"


@pytest.mark.asyncio
async def test_graphql_error(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = {"errors": [{"message": "test"}]}
    with pytest.raises(GitHubGraphQLException, match="test"):
        await github_api.graphql(EXAMPLE_QUERY)

    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/graphql"



# pylint: disable=missing-docstring, redefined-outer-name, unused-import

import json

import pytest

from aiogithubapi import AIOGitHubAPINotModifiedException, GitHub

from tests.common import TOKEN
from tests.legacy.responses.issue_comments import issue_comments_response
from tests.legacy.responses.issue_fixture import issue_response, issues_response
from tests.legacy.responses.repository_fixture import repository_response


@pytest.mark.asyncio
async def test_issue(aresponses, repository_response, issue_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )

    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        assert issue.html_url == "https://github.com/octocat/Hello-World/issues/1347"
        assert issue.number == 1347
        assert issue.labels == [
            {
                "id": 208045946,
                "node_id": "MDU6TGFiZWwyMDgwNDU5NDY=",
                "url": "https://api.github.com/repos/octocat/Hello-World/labels/bug",
                "name": "bug",
                "description": "Something isn't working",
                "color": "f29513",
                "default": True,
            }
        ]
        assert issue.title == "Found a bug"
        assert issue.state == "open"
        assert issue.assignees == [
            {
                "login": "octocat",
                "id": 1,
                "node_id": "MDQ6VXNlcjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/users/octocat",
                "html_url": "https://github.com/octocat",
                "followers_url": "https://api.github.com/users/octocat/followers",
                "following_url": "https://api.github.com/users/octocat/following{/other_user}",
                "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
                "organizations_url": "https://api.github.com/users/octocat/orgs",
                "repos_url": "https://api.github.com/users/octocat/repos",
                "events_url": "https://api.github.com/users/octocat/events{/privacy}",
                "received_events_url": "https://api.github.com/users/octocat/received_events",
                "type": "User",
                "site_admin": False,
            }
        ]
        assert issue.body == "I'm having a problem with this."
        user = issue.user
        assert user.login == "octocat"
        assert user.id == 1
        assert user.avatar_url == "https://github.com/images/error/octocat_happy.gif"
        assert user.html_url == "https://github.com/octocat"
        assert user.type == "User"
        assert not user.site_admin


@pytest.mark.asyncio
async def test_update_issue(aresponses, repository_response, issue_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "post",
        aresponses.Response(
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        data = {
            "title": "test",
            "body": "body",
            "state": "closed",
            "milestone": "v1.0",
            "labels": ["test"],
            "assignees": ["octocat"],
        }
        await issue.update(**data)


@pytest.mark.asyncio
async def test_get_issue_comments(
    aresponses, repository_response, issue_response, issue_comments_response
):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1/comments",
        "get",
        aresponses.Response(
            text=json.dumps(issue_comments_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        comments = await issue.get_comments()
        assert comments[0].body == "Me too"


@pytest.mark.asyncio
async def test_comment_on_issue(aresponses, repository_response, issue_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1/comments",
        "post",
        aresponses.Response(
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        await issue.comment("test")


@pytest.mark.asyncio
async def test_issue_comments(
    aresponses, repository_response, issue_response, issue_comments_response
):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1/comments",
        "get",
        aresponses.Response(
            text=json.dumps(issue_comments_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        comments = await issue.get_comments()
        first = comments[0]
        assert first.html_url == "https://github.com/octocat/Hello-World/issues/1347#issuecomment-1"
        assert first.body == "Me too"
        assert first.created_at == "2011-04-14T16:00:49Z"
        assert first.updated_at == "2011-04-14T16:00:49Z"
        user = first.user
        assert user.login == "octocat"
        assert user.id == 1
        assert user.avatar_url == "https://github.com/images/error/octocat_happy.gif"
        assert user.html_url == "https://github.com/octocat"
        assert user.type == "User"
        assert not user.site_admin


@pytest.mark.asyncio
async def test_update_issue_comment(
    aresponses, repository_response, issue_response, issue_comments_response
):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1",
        "get",
        aresponses.Response(
            text=json.dumps(issue_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/1/comments",
        "get",
        aresponses.Response(
            text=json.dumps(issue_comments_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues/comments/1",
        "post",
        aresponses.Response(
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issue(1)
        comments = await issue.get_comments()
        await comments[0].update("test")


@pytest.mark.asyncio
async def test_get_issues(aresponses, repository_response, issues_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues",
        "get",
        aresponses.Response(
            text=json.dumps(issues_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/issues",
        "get",
        aresponses.Response(status=304),
    )

    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        issue = await repository.get_issues()
        assert len(issue) == 1

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await repository.get_issues(etag=github.client.last_response.etag)



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import datetime
import json

import pytest

from aiogithubapi import AIOGitHubAPINotModifiedException, GitHub

from tests.common import TOKEN
from tests.legacy.responses.contents import contents_file_response
from tests.legacy.responses.repository_fixture import repository_response
from tests.legacy.responses.tree import tree_response


@pytest.mark.asyncio
async def test_tree_content(aresponses, repository_response, tree_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/git/trees/main",
        "get",
        aresponses.Response(
            text=json.dumps(tree_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/git/trees/main",
        "get",
        aresponses.Response(status=304),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        contents = await repository.get_tree("main")
        assert len([x for x in contents if x.is_directory]) == 1
        assert contents[0].full_path == "subdir/file.txt"
        assert not contents[0].is_directory
        assert contents[0].path == "subdir"
        assert contents[0].filename == "file.txt"
        assert (
            contents[0].url
            == "https://api.github.com/repos/octocat/Hello-World/git/7c258a9869f33c1e1e1f74fbb32f07c86cb5a75b"
        )
        assert (
            contents[0].download_url
            == "https://raw.githubusercontent.com/octocat/Hello-World/main/subdir/file.txt"
        )

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await repository.get_tree("main", etag=github.client.last_response.etag)


@pytest.mark.asyncio
async def test_content(aresponses, repository_response, contents_file_response):
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World",
        "get",
        aresponses.Response(
            text=json.dumps(repository_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/contents/README.md",
        "get",
        aresponses.Response(
            text=json.dumps(contents_file_response),
            status=200,
            headers={
                "X-RateLimit-Remaining": "1337",
                "Content-Type": "application/json",
                "Etag": "xyz..zyx",
            },
        ),
    )
    aresponses.add(
        "api.github.com",
        "/repos/octocat/Hello-World/contents/README.md",
        "get",
        aresponses.Response(status=304),
    )
    async with GitHub(TOKEN) as github:
        repository = await github.get_repo("octocat/Hello-World")
        content = await repository.get_contents("README.md", "main")
        assert content.type == "file"
        assert content.encoding == "base64"
        assert content.name == "README.md"
        assert content.path == "README.md"
        assert content.content == ""
        assert (
            content.download_url
            == "https://raw.githubusercontent.com/octokit/octokit.rb/main/README.md"
        )

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await repository.get_contents(
                "README.md", "main", etag=github.client.last_response.etag
            )



"""Test repos namespace."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI, GitHubRepositoryModel
from aiogithubapi.const import HttpContentType

from tests.common import (
    HEADERS,
    HEADERS_TEXT,
    TEST_REPOSITORY_NAME,
    MockedRequests,
    MockResponse,
)


@pytest.mark.asyncio
async def test_get_repository(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.get(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, GitHubRepositoryModel)
    assert response.data.name == "Hello-World"
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/repos/octocat/hello-world"


@pytest.mark.asyncio
async def test_list_commits(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.list_commits(TEST_REPOSITORY_NAME, sha="main", path="README.md", author="awesome_author", since="2021-01-01", until="2021-01-02")
    assert response.status == 200
    assert isinstance(response.data, list)
    assert response.data[0].commit.message == "Merge pull request #6"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/commits"
    )


@pytest.mark.asyncio
async def test_list_tags(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.list_tags(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert response.data[0].name == "v0.1"
    assert response.data[0].commit.sha == "c5b97d5ae6c19d5c5df71a34c7fbeeda2479ccbc"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"] == "https://api.github.com/repos/octocat/hello-world/tags"
    )


@pytest.mark.asyncio
async def test_tarball(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_headers = {**HEADERS, "content-type": HttpContentType.BASE_GZIP}
    response = await github_api.repos.tarball("octocat/hello-world")
    assert isinstance(response.data, bytes)
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/tarball/"
    )
    response = await github_api.repos.tarball("octocat/hello-world", ref="main")
    assert isinstance(response.data, bytes)
    assert mock_requests.called == 2
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/tarball/main"
    )


@pytest.mark.asyncio
async def test_zipball(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_headers = {**HEADERS, "content-type": HttpContentType.BASE_ZIP}
    response = await github_api.repos.zipball("octocat/hello-world")
    assert isinstance(response.data, bytes)
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/zipball/"
    )
    response = await github_api.repos.zipball("octocat/hello-world", ref="main")
    assert isinstance(response.data, bytes)
    assert mock_requests.called == 2
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/zipball/main"
    )


@pytest.mark.asyncio
async def test_readme(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_headers = HEADERS_TEXT

    response = await github_api.repos.readme(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, str)
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/readme/"
    )

    response = await github_api.repos.readme(TEST_REPOSITORY_NAME, dir="test")
    assert response.status == 200
    assert isinstance(response.data, str)
    assert mock_requests.called == 2
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/readme/test"
    )



"""
Methods for the projects namespace

https://docs.github.com/en/rest/reference/projects
"""

from __future__ import annotations

from typing import Any, Dict

from aiogithubapi.helpers import repository_full_name

from ..const import GitHubRequestKwarg, HttpMethod, RepositoryType
from ..models.projects import GitHubProjectModel
from ..models.response import GitHubResponseModel
from .base import BaseNamespace


class _BaseProjectsNamespace(BaseNamespace):
    """Methods for the base projects namespace"""

    async def _list(
        self,
        endpoint: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubProjectModel]]:
        """Internal helper"""
        response = await self._client.async_call_api(endpoint=endpoint, **kwargs)
        response.data = [GitHubProjectModel(project) for project in response.data or []]
        return response

    async def _get(
        self,
        endpoint: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """Internal helper"""
        response = await self._client.async_call_api(endpoint=endpoint, **kwargs)
        response.data = GitHubProjectModel(response.data)
        return response

    async def _update(
        self,
        endpoint: str,
        data: dict[str, Any],
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """Internal helper"""
        response = await self._client.async_call_api(
            endpoint=endpoint,
            data=data,
            method=HttpMethod.PATCH,
            **kwargs,
        )
        response.data = GitHubProjectModel(response.data)
        return response

    async def _create(
        self,
        endpoint: str,
        data: dict[str, Any],
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """Internal helper"""
        response = await self._client.async_call_api(
            endpoint=endpoint,
            data=data,
            method=HttpMethod.POST,
            **kwargs,
        )
        response.data = GitHubProjectModel(response.data)
        return response

    async def _delete(
        self,
        endpoint: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[None]:
        """Internal helper"""
        return await self._client.async_call_api(
            endpoint=endpoint,
            method=HttpMethod.DELETE,
            **kwargs,
        )


class GitHubBaseProjectsNamespace(_BaseProjectsNamespace):
    """Methods for the organization projects namespace"""

    async def get(
        self,
        project_id: int,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """
         Get a project

         **Arguments**:

         `project_id`

         The project ID to return

        https://docs.github.com/en/rest/reference/projects#list-organization-projects
        """
        return await super()._get(endpoint=f"/projects/{project_id}", **kwargs)

    async def update(
        self,
        project_id: int,
        data: dict[str, Any],
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """
         Get a project

         **Arguments**:

         `project_id`

         The project ID to return

         `data`

         A dictionary of the data to update.

        https://docs.github.com/en/rest/reference/projects#update-a-project
        """
        return await super()._update(endpoint=f"/projects/{project_id}", data=data, **kwargs)

    async def delete(
        self,
        project_id: int,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[None]:
        """
         Delete a project

         **Arguments**:

         `project_id`

         The project ID to return

        https://docs.github.com/en/rest/reference/projects#delete-a-project
        """
        return await super()._delete(endpoint=f"/projects/{project_id}", **kwargs)


class GitHubOrganizationProjectsNamespace(_BaseProjectsNamespace):
    """Methods for the organization projects namespace"""

    async def list(
        self,
        org: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubProjectModel]]:
        """
         List organization projects

         **Arguments**:

         `org`

         The organization to return, example "octocat"

        https://docs.github.com/en/rest/reference/projects#list-organization-projects
        """
        return await super()._list(endpoint=f"/orgs/{org}/projects", **kwargs)

    async def create(
        self,
        org: str,
        name: str,
        *,
        body: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """
         List organization projects

         **Arguments**:

         `org`

         The organization to create the project for, example "octocat"

         `name`

         The name of the project.

         `body`

         The description of the project.

        https://docs.github.com/en/rest/reference/projects#create-an-organization-project
        """
        return await super()._create(
            endpoint=f"/orgs/{org}/projects",
            data={"name": name, "body": body},
            **kwargs,
        )


class GitHubRepositoryProjectsNamespace(_BaseProjectsNamespace):
    """Methods for the repository projects namespace"""

    async def list(
        self,
        repository: RepositoryType,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubProjectModel]]:
        """
         List repository projects

         **Arguments**:

         `repository`

         The repository to return the zip from, example "octocat/hello-world"

        https://docs.github.com/en/rest/reference/projects#list-repository-projects
        """
        return await super()._list(
            endpoint=f"/repos/{repository_full_name(repository)}/projects",
            **kwargs,
        )

    async def create(
        self,
        repository: RepositoryType,
        name: str,
        *,
        body: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """
         Create a new repoitory project

         **Arguments**:

         `repository`

         The repository to return the zip from, example "octocat/hello-world"

         `name`

         The name of the project.

         `body`

         The description of the project.

        https://docs.github.com/en/rest/reference/projects#create-an-organization-project
        """
        return await super()._create(
            endpoint=f"/repos/{repository_full_name(repository)}/projects",
            data={"name": name, "body": body},
            **kwargs,
        )


class GitHubUsersProjectsNamespace(_BaseProjectsNamespace):
    """Methods for the users projects namespace"""

    async def list(
        self,
        username: str,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubProjectModel]]:
        """
         List user projects

         **Arguments**:

         `username`

         The username to return project from, example "octocat"

        https://docs.github.com/en/rest/reference/projects#list-user-projects
        """
        return await super()._list(endpoint=f"/users/{username}/projects", **kwargs)


class GitHubUserProjectsNamespace(_BaseProjectsNamespace):
    """Methods for the user projects namespace"""

    async def create(
        self,
        name: str,
        *,
        body: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubProjectModel]:
        """
         Create a new project for the authenticated user

         **Arguments**:

         `name`

         The name of the project.

         `body`

         The description of the project.

        https://docs.github.com/en/rest/reference/projects#create-a-user-project
        """
        return await super()._create(
            endpoint=f"/user/projects",
            data={"name": name, "body": body},
            **kwargs,
        )



"""GitHub permissions data class."""

from __future__ import annotations

from .base import GitHubDataModelBase


class GitHubPermissionsModel(GitHubDataModelBase):
    """GitHub permissions data class."""

    admin: bool | None = None
    maintain: bool | None = None
    push: bool | None = None
    triage: bool | None = None
    pull: bool | None = None



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import pytest

from aiogithubapi import GitHub

from tests.legacy.responses.base import base_response


@pytest.mark.asyncio
async def test_get_rate_limit(mock_response, base_response, github: GitHub):
    mock_response.mock_data = base_response

    rate_limit = await github.get_rate_limit()
    assert rate_limit["remaining"] == "4999"



"""GitHub commit data class."""

from __future__ import annotations

from typing import Any, Dict

from .base import GitHubDataModelBase
from .user import GitHubUserModel


class _Author(GitHubDataModelBase):
    """Internal entry."""

    name: str | None = None
    email: str | None = None
    date: str | None = None


class _Committer(_Author):
    """Internal entry."""


class _Tree(GitHubDataModelBase):
    """Internal entry."""

    sha: str | None = None
    url: str | None = None


class _Verification(GitHubDataModelBase):
    """Internal entry."""

    verified: bool | None = None
    reason: str | None = None
    signature: str | None = None
    payload: str | None = None


class _Parents(GitHubDataModelBase):
    """Internal entry."""

    sha: str | None = None
    url: str | None = None
    html_url: str | None = None


class _Commit(GitHubDataModelBase):
    """Internal entry."""

    author: _Author | None = None
    committer: _Committer | None = None
    message: str | None = None
    tree: _Tree | None = None
    url: str | None = None
    comment_count: int | None = None
    verification: _Verification | None = None

    def _generate_author(self, data: Dict[str, Any] | None) -> _Author:
        """Generate author data."""
        return _Author(data) if data else None

    def _generate_committer(self, data: Dict[str, Any] | None) -> _Committer:
        """Generate committer data."""
        return _Committer(data) if data else None

    def _generate_tree(self, data: Dict[str, Any] | None) -> _Tree:
        """Generate tree data."""
        return _Tree(data) if data else None

    def _generate_verification(self, data: Dict[str, Any] | None) -> _Verification:
        """Generate verification data."""
        return _Verification(data) if data else None


class GitHubCommitModel(GitHubDataModelBase):
    """GitHub commit data class."""

    sha: str | None = None
    commit: _Commit | None = None
    url: str | None = None
    html_url: str | None = None
    comments_url: str | None = None
    author: GitHubUserModel | None = None
    committer: GitHubUserModel | None = None
    parents: list[_Parents] | None = None

    def _generate_commit(self, data: Dict[str, Any] | None) -> _Commit:
        """Generate commit data."""
        return _Commit(data) if data else None

    def _generate_author(self, data: Dict[str, Any] | None) -> GitHubUserModel:
        """Generate author data."""
        return GitHubUserModel(data) if data else None

    def _generate_committer(self, data: Dict[str, Any] | None) -> GitHubUserModel:
        """Generate committer data."""
        return GitHubUserModel(data) if data else None



"""Test notifications namespace."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI
from aiogithubapi.models.notification import GitHubNotificationModel

from tests.common import (
    MockedRequests,
)


@pytest.mark.asyncio
async def test_list_notifications(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.notifications.list()
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 1
    assert isinstance(response.data[0], GitHubNotificationModel)
    assert response.data[0].id == "1"
    assert response.data[0].subject.title == "Greetings"
    assert response.data[0].reason == "subscribed"
    assert response.data[0].unread is True
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/notifications"



"""
Methods for the notifications namespace

https://docs.github.com/en/rest/activity/notifications
"""

from __future__ import annotations

from typing import Any

from ..const import GitHubRequestKwarg
from ..models.notification import GitHubNotificationModel
from ..models.response import GitHubResponseModel
from .base import BaseNamespace


class GitHubNotificationsNamespace(BaseNamespace):
    """Methods for the notifications namespace"""

    async def list(
        self,
        *,
        all: bool = False,
        participating: bool = False,
        since: str | None = None,
        before: str | None = None,
        page: int = 1,
        per_page: int = 50,
        params: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> GitHubResponseModel[list[GitHubNotificationModel]]:
        """
        List notifications for the authenticated user

        https://docs.github.com/en/rest/activity/notifications#list-notifications-for-the-authenticated-user
        """
        request_params = {}
        if all:
            request_params["all"] = "true"
        if participating:
            request_params["participating"] = "true"
        if since:
            request_params["since"] = since
        if before:
            request_params["before"] = before
        if page != 1:
            request_params["page"] = str(page)
        if per_page != 50:
            request_params["per_page"] = str(per_page)
        if params:
            request_params.update(params)

        response = await self._client.async_call_api(
            endpoint="/notifications",
            params=request_params,
            **kwargs,
        )
        response.data = [GitHubNotificationModel(data) for data in response.data]
        return response



"""
Methods for the contents namespace

https://docs.github.com/en/rest/reference/repos#contents
"""

from __future__ import annotations

from typing import Any, Dict

from aiohttp.hdrs import ACCEPT

from ..const import GitHubRequestAcceptHeader, GitHubRequestKwarg, RepositoryType
from ..helpers import repository_full_name
from ..models.contents import GitHubContentsModel
from ..models.response import GitHubResponseModel
from .base import BaseNamespace


class GitHubContentsNamespace(BaseNamespace):
    """Methods for the contents namespace"""

    async def get(
        self,
        repository: RepositoryType,
        path: str | None = None,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubContentsModel | list[GitHubContentsModel]]:
        """
         Get repository content

        Gets the contents of a file or directory in a repository.

        Files and symlinks support a custom media type for retrieving the raw content
        or rendered HTML (when supported). All content types support a custom media
        type to ensure the content is returned in a consistent object format.

         **Arguments**:

         `repository`

         The repository to return contents from, example "octocat/hello-world"

         `path`

         Specify the file path or directory. If you omit this, you will
         receive the contents of the repository's root directory.

        https://docs.github.com/en/rest/reference/repos#get-repository-content
        """
        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}"
            f"/contents{f'/{path}' if path else ''}",
            **{
                GitHubRequestKwarg.HEADERS: {ACCEPT: GitHubRequestAcceptHeader.BASE_JSON.value},
                **kwargs,
            },
        )

        if isinstance(response.data, list):
            response.data = [GitHubContentsModel(data) for data in response.data]
        else:
            response.data = GitHubContentsModel(response.data)
        return response



"""Test contents namespace."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI

from tests.common import TEST_REPOSITORY_NAME, MockedRequests, MockResponse


@pytest.mark.asyncio
async def test_list(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.pulls.list(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 1
    assert response.data[0].title == "Awesome title"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/pulls"
    )


@pytest.mark.asyncio
async def test_list_empty(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = []
    response = await github_api.repos.pulls.list(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 0
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/pulls"
    )



"""
Methods for the issues namespace

https://docs.github.com/en/rest/reference/pulls
"""

from typing import Any, Dict, List

from ..const import GitHubRequestKwarg, RepositoryType
from ..helpers import repository_full_name
from ..models.pull_request import GitHubPullRequestModel
from ..models.response import GitHubResponseModel
from .base import BaseNamespace


class GitHubPullsNamespace(BaseNamespace):
    """
    Methods for the pull requests namespace

    The Pull Request API allows you to list, view, edit, create,
    and even merge pull requests. Comments on pull requests
    can be managed via the issues namespace.
    """

    async def list(
        self,
        repository: RepositoryType,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[List[GitHubPullRequestModel]]:
        """
         List pull requests

         **Arguments**:

         `repository`

         The repository to return pull requests from, example "octocat/hello-world"

        https://docs.github.com/en/rest/reference/pulls#list-pull-requests
        """
        response = await self._client.async_call_api(
            endpoint=f"/repos/{repository_full_name(repository)}/pulls",
            **kwargs,
        )
        response.data = [GitHubPullRequestModel(data) for data in response.data]
        return response



"""AIOGitHubAPI: AIOGitHubAPI"""

from __future__ import annotations

import os
from typing import List, Optional

import aiohttp

from ..common.const import ACCEPT_HEADERS
from ..objects.base import AIOGitHubAPIBase
from ..objects.repository import AIOGitHubAPIRepository
from .client import AIOGitHubAPIClient


class AIOGitHubAPI(AIOGitHubAPIBase):
    """
    AIOGitHubAPI

    This is the main class this is where it all starts.

    DEPRECATED: Use `GitHubAPI` instead.
    """

    _close_session = False

    def __init__(
        self,
        token: str = None,
        session: aiohttp.ClientSession = None,
        headers: Optional[dict] = None,
        base_url: Optional[str] = None,
    ) -> None:
        """
        Initialises a GitHub API client.

        param | required | description
        -- | -- | --
        `token` | False | [Your GitHub Personal Access Token](https://github.com/settings/tokens).
        `session` | False | `aiohttp.ClientSession` to be used by this package.
        `headers` | False | Custom headers to be used by this package.
        `base_url` | False | Base URL of the GitHub API.

        """
        if session is None:
            session = aiohttp.ClientSession()
            self._close_session = True

        if token is None:
            token = os.getenv("GITHUB_TOKEN")

        self.client = AIOGitHubAPIClient(session, token, headers, base_url)

    async def __aenter__(self) -> "AIOGitHubAPI":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self._close()

    async def get_repo(
        self,
        repo: str,
        etag: Optional[str] = None,
    ) -> AIOGitHubAPIRepository:
        """Return AIOGitHubAPIRepository object."""
        _endpoint = f"/repos/{repo}"
        _headers = {"Accept": ACCEPT_HEADERS["preview"]}
        if etag:
            _headers[aiohttp.hdrs.IF_NONE_MATCH] = etag

        response = await self.client.get(endpoint=_endpoint, headers=_headers)

        return AIOGitHubAPIRepository(self.client, response.data)

    async def get_org_repos(
        self, org: str, page: int = 1, etag: Optional[str] = None
    ) -> List[AIOGitHubAPIRepository]:
        """
        Return a list of AIOGitHubAPIRepository objects.

        param | required | Default | description
        -- | -- | -- | --
        `org` | True | None | The name of the organization, example "octocat"
        `page` | False | 1 | The page number you want to fetch.
        """
        _enpoint = f"/orgs/{org}/repos?page={str(page)}"
        _params = {"per_page": 100}
        _headers = {"Accept": ACCEPT_HEADERS["preview"]}
        if etag:
            _headers[aiohttp.hdrs.IF_NONE_MATCH] = etag

        response = await self.client.get(endpoint=_enpoint, params=_params, headers=_headers)

        return [AIOGitHubAPIRepository(self.client, x) for x in response.data or []]

    async def graphql(self, query: str, variables: dict = {}) -> dict:
        response = await self.client.post(
            endpoint="/graphql",
            returnjson=True,
            data={"query": query, "variables": variables},
            jsondata=True,
        )
        return response.data.get("data")

    async def get_rate_limit(self) -> dict:
        """Return current rate limits."""
        _endpoint = "/rate_limit"

        await self.client.get(endpoint=_endpoint)
        return self.client.ratelimits.__dict__

    async def render_markdown(self, content: str, etag: Optional[str] = None) -> str:
        """
        Return AIOGitHubAPIRepository object.

        [API Docs](https://docs.github.com/en/rest/reference/markdown#render-a-markdown-document-in-raw-mode)

        param | description
        -- | --
        `content` | The content (as markdown) you want to render as GHF Markdown
        """
        _endpoint = "/markdown/raw"
        _headers = {"Content-Type": "text/plain"}
        if etag:
            _headers[aiohttp.hdrs.IF_NONE_MATCH] = etag

        response = await self.client.post(endpoint=_endpoint, headers=_headers, data=content)

        return response.data

    async def _close(self) -> None:
        """Close open client session."""
        if self.client.session and self._close_session:
            await self.client.session.close()



"""AIOGitHubAPI: objects.base"""

from __future__ import annotations

from typing import TYPE_CHECKING

from aiohttp.hdrs import ETAG

from ..common.const import HttpStatusCode

if TYPE_CHECKING:
    from ..client import AIOGitHubAPIClient


class AIOGitHubAPIBase:
    """Base class for AIOGitHubAPI."""

    def __init__(self, attributes) -> None:
        """Initialize."""
        self.attributes = attributes


class AIOGitHubAPIBaseClient(AIOGitHubAPIBase):
    """Base class for AIOGitHubAPI."""

    def __init__(self, client: AIOGitHubAPIClient, attributes: dict) -> None:
        """Initialise."""
        super().__init__(attributes)
        self.client = client


class AIOGitHubAPIResponse:
    """Response object for AIOGitHub."""

    def __init__(self) -> None:
        """initialise."""
        self.headers: dict = {}
        self.data = {}
        self.status: HttpStatusCode = HttpStatusCode.OK

    def as_dict(self):
        """Return attributes as a dict."""
        return {
            "headers": self.headers,
            "status": self.status,
            "data": self.data,
            "etag": self.etag,
        }

    @property
    def etag(self):
        """Return the ETag for this response."""
        return self.headers.get(ETAG)



"""Test client construction"""
# pylint: disable=missing-docstring,protected-access
from __future__ import annotations

import asyncio
from unittest.mock import patch

import aiohttp
import pytest

from aiogithubapi import (
    GitHubAPI,
    GitHubConnectionException,
    GitHubException,
    GitHubRatelimitException,
)
from aiogithubapi.client import MESSAGE_EXCEPTIONS, STATUS_EXCEPTIONS
from aiogithubapi.const import GitHubRequestKwarg, HttpStatusCode

from tests.common import (
    EXPECTED_ETAG,
    TEST_REQUEST_HEADERS,
    MockedRequests,
    MockResponse,
)


@pytest.mark.asyncio
async def test_passing_etag_to_request(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
):
    response = await github_api.generic("/generic")
    assert response.status == 200
    assert "If-None-Match" not in mock_requests.last_request["headers"]
    mock_requests.clear()

    response = await github_api.generic("/generic", **{GitHubRequestKwarg.ETAG: "test"})
    assert response.status == 200
    assert mock_requests.last_request["headers"]["If-None-Match"] == "test"


@pytest.mark.asyncio
async def test_passing_headers_to_request(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
):
    response = await github_api.generic("/generic")
    assert response.status == 200
    assert mock_requests.last_request["headers"] == TEST_REQUEST_HEADERS
    mock_requests.clear()

    response = await github_api.generic(
        "/generic", **{GitHubRequestKwarg.HEADERS: {"test": "test"}}
    )
    assert response.status == 200
    assert mock_requests.last_request["headers"] == {
        **TEST_REQUEST_HEADERS,
        "test": "test",
    }


@pytest.mark.asyncio
async def test_passing_params_to_request(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
):
    response = await github_api.generic("/generic")
    assert response.status == 200
    assert mock_requests.last_request["params"] == {}
    mock_requests.clear()

    response = await github_api.generic("/generic", **{GitHubRequestKwarg.PARAMS: {"test": "test"}})
    assert response.status == 200
    assert mock_requests.last_request["params"] == {"test": "test"}
    mock_requests.clear()

    response = await github_api.generic("/generic", **{GitHubRequestKwarg.QUERY: {"test": "test"}})
    assert response.status == 200
    assert mock_requests.last_request["params"] == {"test": "test"}


@pytest.mark.asyncio
async def test_passing_method_to_request(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
):
    response = await github_api.generic("/generic")
    assert response.status == 200
    assert mock_requests.last_request["method"] == "get"
    mock_requests.clear()

    response = await github_api.generic("/generic", **{GitHubRequestKwarg.METHOD: "POST"})
    assert response.status == 200
    assert mock_requests.last_request["method"] == "post"


@pytest.mark.asyncio
async def test_client_exceptions(github_api: GitHubAPI):
    with patch("aiohttp.ClientSession.request", side_effect=aiohttp.ClientError("client_error")):
        with pytest.raises(
            GitHubConnectionException,
            match="Request exception for 'https://api.github.com/generic' with - client_error",
        ):
            await github_api.generic("/generic")

    with patch(
        "aiohttp.ClientSession.request",
        side_effect=asyncio.CancelledError("cancelation_error"),
    ):
        with pytest.raises(
            GitHubConnectionException,
            match="Request exception for 'https://api.github.com/generic' with - cancelation_error",
        ):
            await github_api.generic("/generic")

    with patch("aiohttp.ClientSession.request", side_effect=asyncio.TimeoutError):
        with pytest.raises(
            GitHubConnectionException,
            match="Timeout of 20 reached while waiting for https://api.github.com/generic",
        ):
            await github_api.generic("/generic")

    with patch("aiohttp.ClientSession.request", side_effect=BaseException("Eeeeeh... okey?")):
        with pytest.raises(
            GitHubException,
            match="Unexpected exception for "
            "'https://api.github.com/generic' with - Eeeeeh... okey?",
        ):
            await github_api.generic("/generic")

    with patch("json.loads", side_effect=KeyError("err")):
        with pytest.raises(
            GitHubException,
            match="Could not handle response data from 'https://api.github.com/generic' with - 'err'",
        ):
            await github_api.generic("/generic")


@pytest.mark.parametrize(
    "status,exception",
    (*((status, STATUS_EXCEPTIONS[status]) for status in STATUS_EXCEPTIONS),),
)
@pytest.mark.asyncio
async def test_status_code_handling(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    status: HttpStatusCode,
    exception: GitHubException,
):

    mock_response.mock_status = status
    mock_response.mock_data = {
        "message": "Error",
        "documentation_url": "https://example.com",
    }

    with pytest.raises(exception):
        await github_api.generic("/generic")


@pytest.mark.parametrize(
    "message,exception",
    (
        *((message, MESSAGE_EXCEPTIONS[message]) for message in MESSAGE_EXCEPTIONS),
        ("API rate limit exceeded for xxx.xxx.xxx.xxx.", GitHubRatelimitException),
        ("Random error", GitHubException),
    ),
)
@pytest.mark.asyncio
async def test_error_message_handling(
    github_api: GitHubAPI,
    mock_response: MockResponse,
    message: str,
    exception: GitHubException,
):

    mock_response.mock_data = {
        "message": message,
        "documentation_url": "https://example.com",
    }

    with pytest.raises(exception):
        await github_api.generic("/generic")


@pytest.mark.asyncio
async def test_no_content_handling(github_api: GitHubAPI, mock_response: MockResponse):
    mock_response.mock_status = 204
    response = await github_api.generic("/generic")
    assert response.data is None


@pytest.mark.asyncio
async def test_response_object(github_api: GitHubAPI, mock_response: MockResponse):
    response = await github_api.generic("/generic")
    assert response.etag == EXPECTED_ETAG
    assert response.page_number == 2
    assert response.next_page_number == 3
    assert response.last_page_number == 46
    assert not response.is_last_page

    mock_response.mock_headers = {"Link": None}
    response = await github_api.generic("/generic")
    assert response.is_last_page
    assert response.page_number == 1
    assert response.next_page_number is None
    assert response.last_page_number is None
    assert response.pages == {}

    mock_response.mock_headers = {"Link": ""}
    response = await github_api.generic("/generic")
    assert response.page_number == 1


@pytest.mark.asyncio
async def test_created_with_message_handling(github_api: GitHubAPI, mock_response: MockResponse):
    """Ensure 201 Created with a message does not raise an exception."""
    mock_response.mock_status = 201
    mock_response.mock_data = {
        "message": "Some info message",
        "id": 123,
        "name": "created-object"
    }
    response = await github_api.generic("/generic")
    assert response.status == HttpStatusCode.CREATED
    assert response.data["message"] == "Some info message"
    assert response.data["id"] == 123
    assert response.data["name"] == "created-object"



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
from aiogithubapi.objects.ratelimit import AIOGitHubAPIRateLimit


def test_ratelimit():
    ratelimit = AIOGitHubAPIRateLimit()
    assert ratelimit.reset_utc is None
    ratelimit.load_from_response_headers(
        {
            "X-RateLimit-Remaining": "4999",
            "X-RateLimit-Limit": "5000",
            "X-RateLimit-Reset": "4999",
        }
    )
    assert ratelimit.reset_utc.year == 1970
    assert ratelimit.reset_utc.month == 1
    assert ratelimit.reset_utc.day == 1
    assert ratelimit.reset_utc.hour == 1
    assert ratelimit.reset_utc.minute == 23
    assert ratelimit.reset_utc.second == 19



"""GitHub projects model."""

from __future__ import annotations

from typing import Any, Dict

from .base import GitHubDataModelBase
from .user import GitHubBaseUserModel


class GitHubProjectModel(GitHubDataModelBase):
    """GitHub project model."""

    owner_url: str | None = None
    url: str | None = None
    html_url: str | None = None
    columns_url: str | None = None
    id: int | None = None
    node_id: str | None = None
    name: str | None = None
    body: str | None = None
    number: int | None = None
    state: str | None = None
    creator: GitHubBaseUserModel | None = None
    created_at: str | None = None
    updated_at: str | None = None

    def _generate_creator(self, data: Dict[str, Any] | None) -> GitHubBaseUserModel:
        """Generate GitHub user creator model."""
        return GitHubBaseUserModel(data) if data else None



"""
Class for OAuth device flow authentication.

https://docs.github.com/en/developers/apps/authorizing-oauth-apps#device-flow
"""

from __future__ import annotations

import asyncio
from datetime import datetime

import aiohttp

from ..common.const import (
    LOGGER,
    OAUTH_ACCESS_TOKEN,
    OAUTH_DEVICE_LOGIN,
    DeviceFlowError,
    HttpMethod,
)
from ..common.exceptions import AIOGitHubAPIException
from ..objects.login.device import AIOGitHubAPILoginDevice
from ..objects.login.oauth import AIOGitHubAPILoginOauth
from .helpers import async_call_api

HEADERS = {"Accept": "application/json"}


class AIOGitHubAPIDeviceLogin:
    _close_session = False

    def __init__(
        self,
        client_id: str,
        scope: str = "",
        session: aiohttp.ClientSession = None,
    ):
        """
        Initialises a GitHub API OAuth device flow.

        param | required | description
        -- | -- | --
        `client_id` | True | The client ID of your OAuth app.
        `scope` | False | [Scope(s)](https://docs.github.com/en/developers/apps/scopes-for-oauth-apps) that will be requested.
        `session` | False | `aiohttp.ClientSession` to be used by this package.
        """
        self.client_id = client_id
        self.scope = scope
        self._interval = 5
        self._expires_in = None
        self._expires = None
        self._device_code = None

        if session is None:
            self.session = aiohttp.ClientSession()
            self._close_session = True
        else:
            self.session = session

    async def __aenter__(self) -> "AIOGitHubAPIDeviceLogin":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self._close()

    async def async_register_device(self) -> AIOGitHubAPILoginDevice:
        """Register the device and return an object that contains the user code for authorization."""
        params = {"client_id": self.client_id, "scope": self.scope}
        response = await async_call_api(
            session=self.session,
            method=HttpMethod.POST,
            url=OAUTH_DEVICE_LOGIN,
            params=params,
            headers=HEADERS,
        )
        device = AIOGitHubAPILoginDevice(response.data)
        self._device_code = device.device_code
        self._interval = device.interval
        self._expires = datetime.timestamp(datetime.now()) + device.expires_in

        return device

    async def async_device_activation(self) -> AIOGitHubAPILoginOauth:
        """Wait for the user to enter the code and activate the device."""
        _activation = None
        while _activation is None:
            if self._expires is None or self._device_code is None:
                await asyncio.sleep(self._interval)

            params = {
                "client_id": self.client_id,
                "device_code": self._device_code,
                "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            }

            if self._expires < datetime.timestamp(datetime.now()):
                raise AIOGitHubAPIException("User took too long to enter key")

            try:
                response = await async_call_api(
                    session=self.session,
                    method=HttpMethod.POST,
                    url=OAUTH_ACCESS_TOKEN,
                    params=params,
                    headers=HEADERS,
                )
                if response.data.get("error"):
                    if response.data["error"] == DeviceFlowError.AUTHORIZATION_PENDING:
                        LOGGER.debug(response.data["error_description"])
                        await asyncio.sleep(self._interval)
                    else:
                        raise AIOGitHubAPIException(response.data["error_description"])
                else:
                    _activation = AIOGitHubAPILoginOauth(response.data)
                    break

            except AIOGitHubAPIException as exception:
                raise AIOGitHubAPIException(exception) from exception

        return _activation

    async def _close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()



"""GitHub reaction data class."""

from __future__ import annotations

from .base import GitHubDataModelBase


class GitHubReactionModel(GitHubDataModelBase):
    """GitHub reaction data class."""

    _log_missing: bool = False

    url: str | None = None
    total_count: int | None = None
    thumbs_up: int | None = None
    thumbs_down: int | None = None
    laugh: int | None = None
    hooray: int | None = None
    confused: int | None = None
    heart: int | None = None
    rocket: int | None = None
    eyes: int | None = None

    def __post_init__(self):
        """Initialize attributes."""
        self.thumbs_up = self._raw_data.get("+1", 0)
        self.thumbs_down = self._raw_data.get("-1", 0)



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import json

import aiohttp
import pytest

from aiogithubapi import AIOGitHubAPIException, AIOGitHubAPINotModifiedException, GitHub
from aiogithubapi.common.exceptions import AIOGitHubAPIRatelimitException

from tests.common import TOKEN
from tests.legacy.responses.base import base_response


@pytest.mark.asyncio
async def test_render_markdown(mock_response, base_response, client_session):
    mock_response.mock_data = json.dumps(base_response)

    async with GitHub(TOKEN, session=client_session) as github:
        render = await github.render_markdown("test")
        assert github.client.ratelimits.remaining == "4999"
        assert isinstance(render, str)

        mock_response.clear()
        mock_response.mock_status = 304

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await github.render_markdown("test", etag=github.client.last_response.etag)


@pytest.mark.asyncio
async def test_render_markdown_rate_limited(mock_response, github):
    mock_response.mock_status = 403
    with pytest.raises(AIOGitHubAPIRatelimitException):
        await github.render_markdown("test")


@pytest.mark.asyncio
async def test_render_markdown_error(mock_response, client_session):
    mock_response.mock_status = 500

    async with GitHub(TOKEN, session=client_session) as github:
        with pytest.raises(AIOGitHubAPIException):
            await github.render_markdown("test")



"""Fixtures."""
# pylint: disable=missing-docstring
import pytest


@pytest.fixture()
def commit_response():
    return {
        "url": "https://api.github.com/repos/octocat/Hello-World/commits/6dcb09b5b57875f334f61aebed695e2e4193db5e",
        "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
        "node_id": "MDY6Q29tbWl0NmRjYjA5YjViNTc4NzVmMzM0ZjYxYWViZWQ2OTVlMmU0MTkzZGI1ZQ==",
        "html_url": "https://github.com/octocat/Hello-World/commit/6dcb09b5b57875f334f61aebed695e2e4193db5e",
        "comments_url": "https://api.github.com/repos/octocat/Hello-World/commits/6dcb09b5b57875f334f61aebed695e2e4193db5e/comments",
        "commit": {
            "url": "https://api.github.com/repos/octocat/Hello-World/git/commits/6dcb09b5b57875f334f61aebed695e2e4193db5e",
            "author": {
                "name": "Monalisa Octocat",
                "email": "support@github.com",
                "date": "2011-04-14T16:00:49Z",
            },
            "committer": {
                "name": "Monalisa Octocat",
                "email": "support@github.com",
                "date": "2011-04-14T16:00:49Z",
            },
            "message": "Fix all the bugs",
            "tree": {
                "url": "https://api.github.com/repos/octocat/Hello-World/tree/6dcb09b5b57875f334f61aebed695e2e4193db5e",
                "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
            },
            "comment_count": 0,
            "verification": {
                "verified": False,
                "reason": "unsigned",
                "signature": None,
                "payload": None,
            },
        },
        "author": {
            "login": "octocat",
            "id": 1,
            "node_id": "MDQ6VXNlcjE=",
            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
            "gravatar_id": "",
            "url": "https://api.github.com/users/octocat",
            "html_url": "https://github.com/octocat",
            "followers_url": "https://api.github.com/users/octocat/followers",
            "following_url": "https://api.github.com/users/octocat/following{/other_user}",
            "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
            "organizations_url": "https://api.github.com/users/octocat/orgs",
            "repos_url": "https://api.github.com/users/octocat/repos",
            "events_url": "https://api.github.com/users/octocat/events{/privacy}",
            "received_events_url": "https://api.github.com/users/octocat/received_events",
            "type": "User",
            "site_admin": False,
        },
        "committer": {
            "login": "octocat",
            "id": 1,
            "node_id": "MDQ6VXNlcjE=",
            "avatar_url": "https://github.com/images/error/octocat_happy.gif",
            "gravatar_id": "",
            "url": "https://api.github.com/users/octocat",
            "html_url": "https://github.com/octocat",
            "followers_url": "https://api.github.com/users/octocat/followers",
            "following_url": "https://api.github.com/users/octocat/following{/other_user}",
            "gists_url": "https://api.github.com/users/octocat/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/octocat/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/octocat/subscriptions",
            "organizations_url": "https://api.github.com/users/octocat/orgs",
            "repos_url": "https://api.github.com/users/octocat/repos",
            "events_url": "https://api.github.com/users/octocat/events{/privacy}",
            "received_events_url": "https://api.github.com/users/octocat/received_events",
            "type": "User",
            "site_admin": False,
        },
        "parents": [
            {
                "url": "https://api.github.com/repos/octocat/Hello-World/commits/6dcb09b5b57875f334f61aebed695e2e4193db5e",
                "sha": "6dcb09b5b57875f334f61aebed695e2e4193db5e",
            }
        ],
        "stats": {"additions": 104, "deletions": 4, "total": 108},
        "files": [
            {
                "filename": "file1.txt",
                "additions": 10,
                "deletions": 2,
                "changes": 12,
                "status": "modified",
                "raw_url": "https://github.com/octocat/Hello-World/raw/7ca483543807a51b6079e54ac4cc392bc29ae284/file1.txt",
                "blob_url": "https://github.com/octocat/Hello-World/blob/7ca483543807a51b6079e54ac4cc392bc29ae284/file1.txt",
                "patch": "@@ -29,7 +29,7 @@\n.....",
            }
        ],
    }



"""Test contents namespace."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI, GitHubContentsModel

from tests.common import TEST_REPOSITORY_NAME, MockedRequests


@pytest.mark.asyncio
async def test_get_root(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.contents.get(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 1
    assert response.data[0].name == "README"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/contents"
    )


@pytest.mark.asyncio
async def test_get_readme(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.contents.get(TEST_REPOSITORY_NAME, "README")
    assert response.status == 200
    assert isinstance(response.data, GitHubContentsModel)
    assert response.data.name == "README"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/contents/README"
    )



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import json

import pytest

from aiogithubapi import (
    AIOGitHubAPIAuthenticationException,
    AIOGitHubAPIException,
    AIOGitHubAPIRatelimitException,
    GitHub,
)
from aiogithubapi.const import DEFAULT_USER_AGENT

from tests.common import HEADERS, HEADERS_RATELIMITED, TOKEN
from tests.legacy.responses.base import bad_auth_response, bad_response, base_response


@pytest.mark.asyncio
async def test_get(mock_response, client_session):

    async with GitHub(TOKEN, session=client_session) as github:
        await github.client.get("/")
        assert github.client.ratelimits.remaining == "4999"


@pytest.mark.asyncio
async def test_headers(mock_response, client_session):
    mock_response.mock_headers = HEADERS

    async with GitHub(TOKEN, session=client_session) as github:
        await github.client.get("/")
        assert github.client.headers["User-Agent"] == DEFAULT_USER_AGENT

    async with GitHub(
        TOKEN, headers={"User-Agent": "test/client"}, session=client_session
    ) as github:
        await github.client.get("/")
        assert github.client.headers["User-Agent"] == "test/client"


@pytest.mark.asyncio
async def test_post(mock_response, client_session):
    mock_response.mock_headers = HEADERS

    async with GitHub(session=client_session) as github:
        await github.client.post("/")
        assert github.client.ratelimits.remaining == "4999"


@pytest.mark.asyncio
async def test_post_with_json(mock_response, base_response, client_session):
    mock_response.mock_headers = HEADERS
    mock_response.mock_data = base_response

    async with GitHub(TOKEN, session=client_session) as github:
        await github.client.post("/", data={"test": "test"}, jsondata=True)
        assert github.client.ratelimits.remaining == "4999"


@pytest.mark.asyncio
async def test_get_ratelimited(client, mock_response):
    mock_response.mock_status = 403

    with pytest.raises(AIOGitHubAPIRatelimitException):
        await client.get("/")


@pytest.mark.asyncio
async def test_get_unauthorized(client, mock_response):
    mock_response.mock_status = 401

    with pytest.raises(AIOGitHubAPIAuthenticationException):
        await client.get("/")


@pytest.mark.asyncio
async def test_post_ratelimited(client, mock_response, base_response):
    mock_response.data = json.dumps(base_response)
    mock_response.mock_status = 403
    mock_response.mock_headers = HEADERS_RATELIMITED

    with pytest.raises(AIOGitHubAPIRatelimitException):
        await client.post("/")


@pytest.mark.asyncio
async def test_get_error(mock_response, client_session):
    mock_response.mock_status = 500

    async with GitHub(TOKEN, session=client_session) as github:
        with pytest.raises(AIOGitHubAPIException):
            await github.client.get("/")


@pytest.mark.asyncio
async def test_post_error(mock_response, client_session):
    mock_response.mock_status = 500

    async with GitHub(TOKEN, session=client_session) as github:
        with pytest.raises(AIOGitHubAPIException):
            await github.client.post("/")


@pytest.mark.asyncio
async def test_ok_get_auth_error(mock_response, bad_auth_response, client_session):
    mock_response.mock_data = bad_auth_response

    async with GitHub(TOKEN, session=client_session) as github:
        with pytest.raises(AIOGitHubAPIException):
            await github.client.get("/")


@pytest.mark.asyncio
async def test_ok_get_error(mock_response, bad_response, client_session):
    mock_response.mock_data = bad_response

    async with GitHub(TOKEN, session=client_session) as github:
        with pytest.raises(AIOGitHubAPIException):
            await github.client.get("/")


@pytest.mark.asyncio
async def test_custom_base_url(mock_response, mock_requests, client_session):
    mock_response.mock_headers = HEADERS

    async with GitHub(TOKEN, base_url="http://example.com", session=client_session) as github:
        await github.client.get("/")
        assert github.client.ratelimits.remaining == "4999"

    assert mock_requests.last_request["url"] == "http://example.com/"



"""Test issues namespace."""
import pytest

from aiogithubapi import GitHubAPI, GitHubIssueCommentModel, GitHubIssueModel

from tests.common import (
    TEST_REPOSITORY_NAME,
    MockedRequests,
    MockResponse,
    load_fixture,
)


@pytest.mark.asyncio
async def test_list(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.issues.list(TEST_REPOSITORY_NAME)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 1
    assert response.data[0].title == "Test issue"
    assert not response.data[0].is_pull_request
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues"
    )


@pytest.mark.asyncio
async def test_get(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.get(TEST_REPOSITORY_NAME, 1)
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueModel)
    assert response.data.title == "Test issue"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1"
    )


@pytest.mark.asyncio
async def test_get_pull_request_from_issues_namespace(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = {
        **load_fixture(
            "repos_octocat_hello-world_issues.json",
            asjson=True,
            legacy=False,
        )[0],
        "pull_request": {"url": ""},
    }
    response = await github_api.repos.issues.get(TEST_REPOSITORY_NAME, 1)
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueModel)
    assert response.data.title == "Test issue"
    assert isinstance(response.data.as_dict, dict)
    assert response.data.is_pull_request
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1"
    )


@pytest.mark.asyncio
async def test_create(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.create(TEST_REPOSITORY_NAME, {"title": "test"})
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueModel)
    assert response.data.title == "Test issue"
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "post"
    assert mock_requests.last_request["json"] == {"title": "test"}
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues"
    )


@pytest.mark.asyncio
async def test_update(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.update(TEST_REPOSITORY_NAME, 1, {"title": "test"})
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueModel)
    assert response.data.title == "Test issue"
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "patch"
    assert mock_requests.last_request["json"] == {"title": "test"}
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1"
    )


@pytest.mark.asyncio
async def test_lock(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_status = 204
    response = await github_api.repos.issues.lock(TEST_REPOSITORY_NAME, 1, "spam")
    assert response.status == 204
    assert response.data is None
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "put"
    assert mock_requests.last_request["json"] == {"lock_reason": "spam"}
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1/lock"
    )


@pytest.mark.asyncio
async def test_unlock(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_status = 204
    response = await github_api.repos.issues.unlock(TEST_REPOSITORY_NAME, 1)
    assert response.status == 204
    assert response.data is None
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "delete"
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1/lock"
    )


@pytest.mark.asyncio
async def test_list_comments(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.repos.issues.list_comments(TEST_REPOSITORY_NAME, 1)
    assert response.status == 200
    assert isinstance(response.data, list)
    assert len(response.data) == 1
    assert response.data[0].body == "No logs, no issue!"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1/comments"
    )


@pytest.mark.asyncio
async def test_get_comment(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues_1_comments.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.get_comment(TEST_REPOSITORY_NAME, 1)
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueCommentModel)
    assert response.data.body == "No logs, no issue!"
    assert mock_requests.called == 1
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/comments/1"
    )


@pytest.mark.asyncio
async def test_create_comment(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues_1_comments.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.create_comment(
        TEST_REPOSITORY_NAME, 1, {"title": "test"}
    )
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueCommentModel)
    assert response.data.body == "No logs, no issue!"
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "post"
    assert mock_requests.last_request["json"] == {"title": "test"}
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/1/comments"
    )


@pytest.mark.asyncio
async def test_update_comment(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_data = load_fixture(
        "repos_octocat_hello-world_issues_1_comments.json",
        asjson=True,
        legacy=False,
    )[0]
    response = await github_api.repos.issues.update_comment(
        TEST_REPOSITORY_NAME, 1, {"title": "test"}
    )
    assert response.status == 200
    assert isinstance(response.data, GitHubIssueCommentModel)
    assert response.data.body == "No logs, no issue!"
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "patch"
    assert mock_requests.last_request["json"] == {"title": "test"}
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/comments/1"
    )


@pytest.mark.asyncio
async def test_delete_comment(
    github_api: GitHubAPI,
    mock_requests: MockedRequests,
    mock_response: MockResponse,
):
    mock_response.mock_status = 204
    response = await github_api.repos.issues.delete_comment(TEST_REPOSITORY_NAME, 1)
    assert response.status == 204
    assert response.data is None
    assert mock_requests.called == 1
    assert mock_requests.last_request["method"] == "delete"
    assert (
        mock_requests.last_request["url"]
        == "https://api.github.com/repos/octocat/hello-world/issues/comments/1"
    )



"""AIOGitHubAPI: Exceptions"""

from ..exceptions import (
    GitHubAuthenticationException,
    GitHubException,
    GitHubNotModifiedException,
    GitHubRatelimitException,
)


class AIOGitHubAPIException(GitHubException):
    """
    Raise this when something is off.

    Deprecated: use `aiogithubapi.exceptions.GitHubException` instead
    """


class AIOGitHubAPIRatelimitException(GitHubRatelimitException, AIOGitHubAPIException):
    """
    Raise this when we hit the ratelimit.

    Deprecated: use `aiogithubapi.exceptions.GitHubRatelimitException` instead
    """


class AIOGitHubAPINotModifiedException(GitHubNotModifiedException, AIOGitHubAPIException):
    """
    Raise this when we the content was not modified.

    Deprecated: use `aiogithubapi.exceptions.GitHubNotModifiedException` instead
    """


class AIOGitHubAPIAuthenticationException(GitHubAuthenticationException, AIOGitHubAPIException):
    """
    Raise this when there is an authentication issue.

    Deprecated: use `aiogithubapi.exceptions.GitHubAuthenticationException` instead
    """



"""Construction tests for the device object."""
# pylint: disable=protected-access,missing-function-docstring
from __future__ import annotations

import aiohttp
import pytest

from aiogithubapi import GitHubClientKwarg, GitHubDeviceAPI

from tests.common import CLIENT_ID


@pytest.mark.asyncio
async def test_session_creation():
    device = GitHubDeviceAPI(client_id=CLIENT_ID)
    assert device._session
    assert isinstance(device._session, aiohttp.ClientSession)

    assert not device._session.closed
    await device.close_session()
    assert device._session.closed


@pytest.mark.asyncio
async def test_session_pass():
    async with aiohttp.ClientSession() as session:
        device = GitHubDeviceAPI(client_id=CLIENT_ID, session=session)
        assert device._session is session


@pytest.mark.asyncio
async def test_session_creation_with_enter():
    async with GitHubDeviceAPI(client_id=CLIENT_ID) as device:
        assert device._session
        assert isinstance(device._session, aiohttp.ClientSession)
        assert not device._session.closed

    assert device._session.closed


@pytest.mark.asyncio
async def test_base_url():
    async with GitHubDeviceAPI(client_id=CLIENT_ID) as device:
        assert device._client._base_request_data.base_url == "https://github.com"

    async with GitHubDeviceAPI(
        client_id=CLIENT_ID, **{GitHubClientKwarg.BASE_URL: "https://example.com"}
    ) as device:
        assert device._client._base_request_data.base_url == "https://example.com"



# pylint: disable=missing-docstring, redefined-outer-name, unused-import
import json

import pytest

from aiogithubapi import (
    AIOGitHubAPINotModifiedException,
    AIOGitHubAPIRatelimitException,
    GitHub,
    GitHubAPI,
    GitHubNotModifiedException,
)
from aiogithubapi.const import GitHubRequestKwarg

from tests.common import EXPECTED_ETAG, TOKEN
from tests.legacy.responses.base import base_response
from tests.legacy.responses.repository_fixture import repository_response


@pytest.mark.asyncio
async def test_get_repo(mock_response, repository_response, client_session):
    mock_response.mock_data = repository_response

    async with GitHub(TOKEN, session=client_session) as github:
        repository = await github.get_repo("octocat/Hello-World")
        assert repository.description == "This your first repo!"
        assert github.client.last_response.etag == EXPECTED_ETAG
        assert github.client.ratelimits.remaining == "4999"

        mock_response.clear()
        mock_response.mock_status = 304

        with pytest.raises(AIOGitHubAPINotModifiedException):
            await github.get_repo("octocat/Hello-World", github.client.last_response.etag)


@pytest.mark.asyncio
async def test_get_repo_ratelimited(mock_response, github):
    mock_response.mock_status = 403

    with pytest.raises(AIOGitHubAPIRatelimitException):
        await github.get_repo("octocat/Hello-World")



"""
Methods for the authenticated user namespace

https://docs.github.com/en/rest/reference/users#get-the-authenticated-user
"""

from __future__ import annotations

from typing import Any, Dict

from ..const import GitHubRequestKwarg
from ..models.organization import GitHubOrganizationMinimalModel
from ..models.repository import GitHubRepositoryModel
from ..models.response import GitHubResponseModel
from ..models.user import GitHubAuthenticatedUserModel
from .base import BaseNamespace
from .projects import GitHubUserProjectsNamespace


class GitHubUserNamespace(BaseNamespace):
    """Methods for the user namespace"""

    def __post_init__(self) -> None:
        self._projects = GitHubUserProjectsNamespace(self._client)

    @property
    def projects(self) -> GitHubUserProjectsNamespace:
        """Property to access the users projects namespace"""
        return self._projects

    async def get(
        self,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[GitHubAuthenticatedUserModel]:
        """
         Get the authenticated user


        https://docs.github.com/en/rest/reference/users#get-a-user
        """
        response = await self._client.async_call_api(
            endpoint="/user",
            **kwargs,
        )
        response.data = GitHubAuthenticatedUserModel(response.data)

        return response

    async def starred(
        self,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubRepositoryModel]]:
        """
         Get the authenticated user starred repositories


        https://docs.github.com/en/rest/reference/users#get-a-user
        """
        response = await self._client.async_call_api(
            endpoint="/user/starred",
            **kwargs,
        )

        response.data = [GitHubRepositoryModel(data) for data in response.data]

        return response

    async def repos(
        self,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubRepositoryModel]]:
        """
         Get the repositories for the authenticated user


        https://docs.github.com/en/rest/reference/repos#list-repositories-for-a-user
        """
        response = await self._client.async_call_api(
            endpoint="/user/repos",
            **kwargs,
        )

        response.data = [GitHubRepositoryModel(data) for data in response.data]

        return response

    async def orgs(
        self,
        **kwargs: Dict[GitHubRequestKwarg, Any],
    ) -> GitHubResponseModel[list[GitHubOrganizationMinimalModel]]:
        """
         List public organization memberships for the specified user.


        https://docs.github.com/en/rest/reference/orgs#list-organizations-for-the-authenticated-user
        """
        response = await self._client.async_call_api(endpoint="/user/orgs", **kwargs)
        response.data = [GitHubOrganizationMinimalModel(data) for data in response.data or []]
        return response



"""GitHub license data class."""

from __future__ import annotations

from .base import GitHubDataModelBase


class GitHubLicenseModel(GitHubDataModelBase):
    """GitHub license data class."""

    key: str | None = None
    name: str | None = None
    spdx_id: str | None = None
    url: str | None = None



"""
Generated by generate/generate.py - 2020-08-02 10:32:07.005845
"""
from aiogithubapi.objects.repos.label import AIOGitHubAPIReposLabel

from tests.legacy.responses.repos.label_fixtrue import label_fixtrue_response


def test_label(label_fixtrue_response):
    obj = AIOGitHubAPIReposLabel(label_fixtrue_response)
    assert obj.id == label_fixtrue_response["id"]
    assert obj.node_id == label_fixtrue_response["node_id"]
    assert obj.url == label_fixtrue_response["url"]
    assert obj.name == label_fixtrue_response["name"]
    assert obj.description == label_fixtrue_response["description"]
    assert obj.color == label_fixtrue_response["color"]
    assert obj.default == label_fixtrue_response["default"]



"""Test emojis."""
# pylint: disable=missing-docstring
import pytest

from aiogithubapi import GitHubAPI

from tests.common import MockedRequests


@pytest.mark.asyncio
async def test_emojis(github_api: GitHubAPI, mock_requests: MockedRequests):
    response = await github_api.emojis()
    assert response.status == 200
    assert isinstance(response.data, dict)
    assert "beers" in response.data
    assert mock_requests.called == 1
    assert mock_requests.last_request["url"] == "https://api.github.com/emojis"



"""Example GrapQL Query to get repository tags."""

EXAMPLE_QUERY = """
query ($owner: String!, $repository: String!, $number_of_tags: Int = 1) {
  repository(owner: $owner, name: $repository) {
    refs(
      refPrefix: "refs/tags/"
      first: $number_of_tags
      orderBy: {field: TAG_COMMIT_DATE, direction: DESC}
    ) {
      tags: nodes {
        name
      }
    }
  }
}
"""
